"""
This file implements the tool `xdsl-tblgen`.

The primary function of `xdsl-tblgen` is to convert dialects defined in mlir
tablegen to python code which defines an xdsl dialect. To import tablegen into
python, it must first be converted to json using the tool `llvm-tblgen` with the
`--dump-json flag`. The generated json should then be input to `xdsl-tblgen` via
stdin or an input file.

Example:
Suppose ${LLVM} is the filepath for the llvm repository with `llvm-tblgen`
built. A test dialect can then be generated by:
```bash
${LLVM}/bin/llvm-tblgen -D test ${LLVM}/mlir/test/tblgen-to-irdl/TestDialect.td -I ${LLVM}/mlir/include --dump-json | xdsl-tblgen
```
In the above, "test" is the name of the dialect to generate, the first file is
the location of the tablegen file, and the folder after `-I` contains common
tablegen definition which the dialect depends on.
"""

import argparse
import json
import subprocess
import textwrap
from dataclasses import dataclass, field
from enum import Enum
from io import StringIO
from keyword import iskeyword
from sys import stdin
from typing import IO, Any


@dataclass
class TblgenRecord:
    js: Any

    def __getitem__(self, i: str) -> Any:
        return self.js[i]

    @property
    def name(self) -> str:
        return self["!name"]

    @property
    def summary(self) -> str:
        return self["summary"]

    def print_summary(self):
        return "" if self.summary == "" else f'"""{self.summary}"""'

    @property
    def superclasses(self) -> set[str]:
        return set(self["!superclasses"])

    @property
    def dialect(self) -> str:
        return self["dialect"]["def"]


class TblgenOp(TblgenRecord):
    @property
    def dialect(self) -> str:
        return self["opDialect"]["def"]

    @property
    def op_name(self) -> str:
        return self["opName"]

    @property
    def assembly_format(self) -> str | None:
        if "assemblyFormat" in self.js:
            assembly = self["assemblyFormat"]
            if isinstance(assembly, str):
                return assembly
        return None

    def _dag_to_py(self, dag: str) -> tuple[tuple[str, str], ...]:
        args = self[dag]["args"]
        return tuple((a["def"], n) for a, n in args)

    @property
    def arguments(self) -> tuple[tuple[str, str], ...]:
        return self._dag_to_py("arguments")

    @property
    def results(self) -> tuple[tuple[str, str], ...]:
        return self._dag_to_py("results")

    @property
    def regions(self) -> tuple[tuple[str, str], ...]:
        return self._dag_to_py("regions")

    @property
    def successors(self) -> tuple[tuple[str, str], ...]:
        return self._dag_to_py("successors")


class TblgenType(TblgenRecord):
    @property
    def type_name(self) -> str:
        return self["typeName"]


class TblgenAttr(TblgenRecord):
    @property
    def attr_name(self) -> str:
        return self["attrName"]


@dataclass
class TblgenLoader:
    """
    Class for converting a json generated by `llvm-tblgen --dump-json` to python
    code which builds an xDSL dialect. Must be initialised with json represented
    as a python object. The generated code is stored as strings in the
    `attributes` and `operations` dictionaries.
    """

    js: Any

    attributes: dict[str, str] = field(default_factory=dict[str, str])
    operations: dict[str, str] = field(default_factory=dict[str, str])
    used_records: set[str] = field(default_factory=set[str])
    anon_counter: int = field(default_factory=int)

    def _get_op(self, name: str) -> TblgenOp:
        self.used_records.add(name)
        return TblgenOp(self.js[name])

    def _get_type(self, name: str) -> TblgenType:
        self.used_records.add(name)
        return TblgenType(self.js[name])

    def _get_attr(self, name: str) -> TblgenAttr:
        self.used_records.add(name)
        return TblgenAttr(self.js[name])

    def _get_record(self, name: str) -> TblgenRecord:
        self.used_records.add(name)
        return TblgenRecord(self.js[name])

    def get_dialect_name(self, dialect: str) -> str:
        self.used_records.add(dialect)
        return self.js[dialect]["name"]

    def generate_dialect(self, tblgen_dialect: str):
        """
        Generate a dialect from the json object, generating all its contained
        operations, types, and attributes and generating python code which
        is stored in this class' fields.
        """
        dialect_name = self.get_dialect_name(tblgen_dialect)

        # Get types
        all_types = self.js["!instanceof"]["TypeDef"]
        for t in all_types:
            ty = self._get_type(t)
            if ty.dialect == tblgen_dialect:
                self.generate_type(ty)

        # Get attributes
        all_attrs = self.js["!instanceof"]["AttrDef"]
        for a in all_attrs:
            attr = self._get_attr(a)
            if attr.dialect == tblgen_dialect:
                self.generate_attr(attr)

        # Get ops
        all_ops = self.js["!instanceof"]["Op"]
        for o in all_ops:
            op = self._get_op(o)
            if op.dialect == tblgen_dialect:
                self.generate_op(op, dialect_name)

    def generate_type(self, tblgen_type: TblgenType):
        """
        Generate a type from the json object, storing python code for it in
        `self.attributes`.
        """

        string = textwrap.dedent(f"""
        @irdl_attr_definition
        class {tblgen_type.name}(ParametrizedAttribute, TypeAttribute):
           {tblgen_type.print_summary()}
           name = "{tblgen_type.type_name}"
        """)

        self.attributes[tblgen_type.name] = string

    def generate_attr(self, tblgen_attr: TblgenAttr):
        """
        Generate an attribute from the json object, storing python code for it in
        `self.attributes`.
        """

        string = textwrap.dedent(f"""
        @irdl_attr_definition
        class {tblgen_attr.name}(ParametrizedAttribute):
           {tblgen_attr.print_summary()}
           name = "{tblgen_attr.attr_name}"
        """)

        self.attributes[tblgen_attr.name] = string

    class _ArgType(Enum):
        SINGLE = 0
        VARIADIC = 1
        OPTIONAL = 2
        PROP = 3
        OPTIONAL_PROP = 4

    def _resolve_type_constraint(self, rec: TblgenRecord | str) -> str:
        if isinstance(rec, str):
            rec = self._get_record(rec)
        if rec.name in self.attributes:
            return f"BaseAttr({rec.name})"

        # match specific types
        match rec.name:
            case "NoneType":
                return "EqAttrConstraint(NoneType())"
            case "AnyInteger":
                return "BaseAttr(IntegerType)"
            case "AnySignlessInteger":
                return textwrap.dedent("""
                ParamAttrConstraint(
                    IntegerType,
                    (AnyAttr(), EqAttrConstraint(SignednessAttr(Signedness.SIGNLESS))),
                )""")
            case "AnySignedInteger":
                return textwrap.dedent("""
                ParamAttrConstraint(
                    IntegerType,
                    (AnyAttr(), EqAttrConstraint(SignednessAttr(Signedness.SIGNED))),
                )
                """)
            case "AnyUnsignedInteger":
                return textwrap.dedent("""
                ParamAttrConstraint(
                    IntegerType,
                    (AnyAttr(), EqAttrConstraint(SignednessAttr(Signedness.UNSIGNED))),
                )
                """)
            case "Index":
                return "EqAttrConstraint(IndexType())"
            case "F16":
                return "EqAttrConstraint(Float16Type())"
            case "F32":
                return "EqAttrConstraint(Float32Type())"
            case "F64":
                return "EqAttrConstraint(Float64Type())"
            case "F80":
                return "EqAttrConstraint(Float80Type())"
            case "F128":
                return "EqAttrConstraint(Float128Type())"
            case "BF16":
                return "EqAttrConstraint(BFloat16Type())"
            case "AnyFloat":
                return "AnyFloatConstr"
            case "AnyComplex":
                return "BaseAttr(ComplexType)"
            case _:
                pass
        if "AnyTypeOf" in rec.superclasses:
            return textwrap.dedent(f"""
            AnyOf(
                (
                    {",".join(self._resolve_type_constraint(x["def"]) for x in rec["allowedTypes"])}
                )
            )
            """)

        if "AllOfType" in rec.superclasses:
            return textwrap.dedent(f"""
            AllOf(
                (
                    {",".join(self._resolve_type_constraint(x["def"]) for x in rec["allowedTypes"])}
                )
            )
            """)

        if "AnyI" in rec.superclasses:
            return textwrap.dedent(f"""
            ParamAttrConstraint(
                IntegerType,
                (EqAttrConstraint(IntAttr({rec["bitwidth"]})), AnyAttr()),
            )
            """)

        if "I" in rec.superclasses:
            return f"EqAttrConstraint(IntegerType({rec['bitwidth']}))"
        if "SI" in rec.superclasses:
            return (
                f"EqAttrConstraint(IntegerType({rec['bitwidth']}, Signedness.SIGNED))"
            )
        if "UI" in rec.superclasses:
            return (
                f"EqAttrConstraint(IntegerType({rec['bitwidth']}, Signedness.UNSIGNED))"
            )
        if "Complex" in rec.superclasses:
            return textwrap.dedent(f"""
            ParamAttrConstraint(
                ComplexType,
                ({self._resolve_type_constraint(rec["elementType"]["def"])},),
            )
            """)

        return "AnyAttr()"

    def _resolve_prop_constraint(self, rec: TblgenRecord | str) -> str:
        if isinstance(rec, str):
            rec = self._get_record(rec)

        if rec.name in self.attributes:
            return f"BaseAttr({rec.name})"

        match rec.name:
            case "BoolAttr":
                return "BaseAttr(BoolAttr)"
            case "IndexAttr":
                return textwrap.dedent("""
                IntegerAttr.constr(IndexTypeConstr)
                """)

            case "APIntAttr":
                return textwrap.dedent("""
                IntegerAttr.constr()
                """)  # TODO can't represent APInt properly

            case "StrAttr":
                return "BaseAttr(StringAttr)"
            case "SymbolNameAttr":
                return "SymbolNameConstraint()"
            case "UnitAttr":
                return "EqAttrConstraint(UnitAttr())"
            case _:
                pass

        if "AnyAttrOf" in rec.superclasses:
            return textwrap.dedent(f"""
            AnyOf(
                [{",".join(self._resolve_prop_constraint(x["def"]) for x in rec["allowedAttributes"])}]
            )
            """)

        if (
            "IntegerAttrBase" in rec.superclasses
            or "SignlessIntegerAttrBase" in rec.superclasses
            or "SignedIntegerAttrBase" in rec.superclasses
            or "UnsignedIntegerAttrBase" in rec.superclasses
        ):
            return textwrap.dedent(f"""
            IntegerAttr.constr({self._resolve_type_constraint(rec["valueType"]["def"])})
            """)

        if "FloatAttrBase" in rec.superclasses:
            return textwrap.dedent(f"""
            ParamAttrConstraint(
                FloatAttr,
                (
                    AnyAttr(),
                    {self._resolve_type_constraint(rec["valueType"]["def"])},
                ),
            )
            """)

        return "AnyAttr()"

    def _resolve_name(self, name: Any) -> str:
        if isinstance(name, str):
            if iskeyword(name):
                return f"{name}_"
            return name

        self.anon_counter += 1
        return f"v{self.anon_counter}"

    def _resolve_constraint(self, rec: TblgenRecord | str) -> tuple[_ArgType, str]:
        if isinstance(rec, str):
            rec = self._get_record(rec)

        superclasses = rec.superclasses
        if "Variadic" in superclasses:
            return (
                self._ArgType.VARIADIC,
                self._resolve_type_constraint(rec["baseType"]["def"]),
            )
        elif "Optional" in superclasses:
            return (
                self._ArgType.OPTIONAL,
                self._resolve_type_constraint(rec["baseType"]["def"]),
            )
        elif "Type" in superclasses:
            return (self._ArgType.SINGLE, self._resolve_type_constraint(rec))
        elif "OptionalAttr" in superclasses:
            return (
                self._ArgType.OPTIONAL_PROP,
                self._resolve_prop_constraint(rec["baseAttr"]["def"]),
            )
        else:
            return (self._ArgType.PROP, self._resolve_prop_constraint(rec))

    def generate_op(self, tblgen_op: TblgenOp, dialect_name: str):
        """
        Generate an operation from the json object, storing python code for it in
        `self.operations`.
        """

        fields = {"name": f'"{dialect_name}.{tblgen_op.op_name}"'}

        assembly = tblgen_op.assembly_format
        if assembly is not None and "custom" not in assembly:
            fields["assembly_format"] = '"""' + assembly + '"""'

        for [arg, orig_name] in tblgen_op.arguments:
            name = self._resolve_name(orig_name)
            (variadicity, constraint) = self._resolve_constraint(arg)
            match variadicity:
                case self._ArgType.SINGLE:
                    fields[name] = f"operand_def({constraint})"
                case self._ArgType.OPTIONAL:
                    fields[name] = f"opt_operand_def({constraint})"
                case self._ArgType.VARIADIC:
                    fields[name] = f"var_operand_def({constraint})"
                case self._ArgType.PROP:
                    name_str = (
                        f', prop_name = "{orig_name}"' if iskeyword(orig_name) else ""
                    )
                    fields[name] = f"prop_def({constraint}{name_str})"
                case self._ArgType.OPTIONAL_PROP:
                    name_str = (
                        f', prop_name = "{orig_name}"' if iskeyword(orig_name) else ""
                    )
                    fields[name] = f"opt_prop_def({constraint}{name_str})"

        for [res, name] in tblgen_op.results:
            name = self._resolve_name(name)
            (variadicity, constraint) = self._resolve_constraint(res)
            match variadicity:
                case self._ArgType.SINGLE:
                    fields[name] = f"result_def({constraint})"
                case self._ArgType.OPTIONAL:
                    fields[name] = f"opt_result_def({constraint})"
                case self._ArgType.VARIADIC:
                    fields[name] = f"var_result_def({constraint})"
                case _:
                    continue

        for [region, name] in tblgen_op.regions:
            name = self._resolve_name(name)
            region = self._get_record(region)
            variadic = "VariadicRegion" in region.superclasses
            single_block = (
                "SizedRegion" in region.superclasses
                and region.summary == "region with 1 blocks"
            )
            match (variadic, single_block):
                case (False, False):
                    fields[name] = "region_def()"
                case (False, True):
                    fields[name] = 'region_def("single_block")'
                case (True, False):
                    fields[name] = "var_region_def()"
                case (True, True):
                    fields[name] = 'var_region_def("single_block")'

        for [succ, name] in tblgen_op.successors:
            name = self._resolve_name(name)
            succ = self._get_record(succ)
            if "VariadicRegion" in succ.superclasses:
                fields[name] = "var_successor_def()"
            else:
                fields[name] = "successor_def()"

        field_string = textwrap.indent(
            "\n\n".join(f"{x} = {d}" for x, d in fields.items()), "    "
        )
        string = f"""
@irdl_op_definition
class {tblgen_op.name}(IRDLOperation):
    {tblgen_op.print_summary()}

{field_string}
"""

        self.operations[tblgen_op.name] = string


def cull_json(output_file: IO[str] | None, loader: TblgenLoader):
    js = loader.js
    required_fields = {
        "!name",
        "!superclasses",
        "assemblyFormat",
        "summary",
        "dialect",
        "opDialect",
        "typeName",
        "attrName",
        "opName",
        "arguments",
        "results",
        "regions",
        "successors",
        "allowedTypes",
        "bitwidth",
        "elementType",
        "valueType",
        "baseType",
        "baseAttr",
        "def",
        "name",
        "allowedAttributes",
    }

    def cull_field(js_in: dict[str, Any]) -> dict[str, Any]:
        return {key: js_in[key] for key in js_in if key in required_fields}

    culled: dict[str, Any] = {key: cull_field(js[key]) for key in loader.used_records}
    culled["!instanceof"] = {
        key: js["!instanceof"][key] for key in ("TypeDef", "AttrDef", "Op", "Dialect")
    }

    print(json.dumps(culled, sort_keys=True, indent=2), file=output_file)


def tblgen_to_dialect(
    input_file: str | None,
    tblgen_dialect: str,
    output_file: IO[str] | None,
    loader: TblgenLoader,
):
    dialect = loader.get_dialect_name(tblgen_dialect)
    with StringIO() as out_str:
        in_file = "stdin" if input_file is None else input_file
        print(
            textwrap.dedent(f"""\
            \"""
            This file is automatically generated by xDSL and not meant to be modified.

            It was generated from {in_file}
            \"""

            # ruff: noqa: F403, F405

            from xdsl.dialects.builtin import *
            from xdsl.ir import *
            from xdsl.irdl import *
            """),
            file=out_str,
        )

        for attr in loader.attributes.values():
            print(attr, file=out_str)

        for op in loader.operations.values():
            print(op, file=out_str)

        print(
            textwrap.dedent(f"""\
            {tblgen_dialect} = Dialect(
                "{dialect}",
                [{" ".join(f"{key}," for key in loader.operations.keys())}],
                [{" ".join(f"{key}," for key in loader.attributes.keys())}],
            )"""),
            file=out_str,
        )

        content = out_str.getvalue()

    # Format output
    output = subprocess.run(
        [
            "ruff",
            "format",
            "--stdin-filename",
            f"{dialect}.py",
        ],
        input=content,
        capture_output=True,
        text=True,
    )

    if output.stderr:
        raise Exception(f"Formatting failed: {output.stderr}")

    print(output.stdout, file=output_file, end="")


def tblgen_to_py(
    input_file: str | None, output: IO[str] | None = None, *, cull: bool = False
):
    if input_file is None:
        in_file = stdin
    else:
        in_file = open(input_file)

    with in_file as file:
        js = json.load(file)
    loader = TblgenLoader(js)
    dialects = js["!instanceof"]["Dialect"]
    [dialect] = dialects
    loader.generate_dialect(dialect)

    if cull:
        cull_json(output, loader)
    else:
        tblgen_to_dialect(input_file, dialect, output, loader)


def main():
    # Parse CLI arguments
    arg_parser = argparse.ArgumentParser(
        description="Convert tblgen json to a Python definition of a xDSL dialect."
    )
    arg_parser.add_argument(
        "-o", "--output-file", required=False, type=str, help="path to output file"
    )
    arg_parser.add_argument(
        "-i", "--input_file", required=False, type=str, help="path to input file"
    )
    arg_parser.add_argument(
        "-c",
        "--cull",
        action="store_true",
        help="Output a culled json with only necessary fields",
    )
    args = arg_parser.parse_args()

    if args.output_file is not None:
        with open(args.output_file, "w") as output:
            tblgen_to_py(args.input_file, output, cull=args.cull)
    else:
        tblgen_to_py(args.input_file, cull=args.cull)



# RUN: python %s | filecheck %s

from xdsl.dialects.arith import AddfOp, MulfOp
from xdsl.dialects.builtin import f64
from xdsl.frontend.pyast.context import CodeContext
from xdsl.frontend.pyast.program import FrontendProgram

p1 = FrontendProgram()
with CodeContext(p1):
    # CHECK: builtin.module {
    # CHECK-NEXT: }
    pass

p1.compile(desymref=False)
print(p1.textual_format())


p2 = FrontendProgram()
p2.register_type(float, f64)
p2.register_function(float.__add__, AddfOp)
p2.register_function(float.__mul__, MulfOp)
with CodeContext(p2):
    # CHECK:       builtin.module {
    # CHECK-NEXT:  func.func @foo(%x : f64, %y : f64, %z : f64) -> f64 {
    # CHECK-NEXT:    %0 = arith.mulf %y, %z : f64
    # CHECK-NEXT:    %1 = arith.addf %x, %0 : f64
    # CHECK-NEXT:    func.return %1 : f64
    # CHECK-NEXT:  }
    # CHECK-NEXT:}
    def foo(x: float, y: float, z: float) -> float:
        return x + y * z


p2.compile()
print(p2.textual_format())



import os
from dataclasses import dataclass

from xdsl.context import Context
from xdsl.dialects import builtin, pdl
from xdsl.interpreters.pdl import (
    PDLRewritePattern,
)
from xdsl.parser import Parser
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriteWalker,
    RewritePattern,
)


@dataclass(frozen=True)
class ApplyPDLPass(ModulePass):
    name = "apply-pdl"

    pdl_file: str | None = None

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        if self.pdl_file is not None:
            assert os.path.exists(self.pdl_file)
            with open(self.pdl_file) as f:
                pdl_module_str = f.read()
                parser = Parser(ctx, pdl_module_str)
                pdl_module = parser.parse_module()
        else:
            pdl_module = op
        rewrite_patterns: list[RewritePattern] = [
            PDLRewritePattern(op, ctx, None)
            for op in pdl_module.walk()
            if isinstance(op, pdl.RewriteOp)
        ]
        pattern_applier = GreedyRewritePatternApplier(rewrite_patterns)
        PatternRewriteWalker(pattern_applier).rewrite_module(op)



"""
This test file needs the other files in the same folder to read and reprint them in the
test functions below.
"""

import re
from contextlib import redirect_stdout
from io import StringIO
from typing import IO

import pytest

from xdsl.context import Context
from xdsl.dialects import builtin, get_all_dialects
from xdsl.passes import ModulePass
from xdsl.transforms import get_all_passes
from xdsl.utils.exceptions import DiagnosticException, ParseError
from xdsl.xdsl_opt_main import xDSLOptMain


def test_dialects_and_passes():
    assert get_all_dialects()
    assert get_all_passes()


def test_opt():
    opt = xDSLOptMain(args=[])
    assert opt.available_frontends.keys()
    assert opt.available_targets.keys()
    assert opt.available_passes.keys()


def test_empty_program():
    filename = "tests/xdsl_opt/empty_program.mlir"
    opt = xDSLOptMain(args=[filename])

    f = StringIO("")
    with redirect_stdout(f):
        opt.run()

    with open(filename) as file:
        expected = file.read()
        assert f.getvalue().strip() == expected.strip()


@pytest.mark.parametrize(
    "args, error_type, expected_error",
    [
        (
            ["--no-implicit-module", "tests/xdsl_opt/not_module.mlir"],
            ParseError,
            "builtin.module operation expected",
        ),
        (
            ["--no-implicit-module", "tests/xdsl_opt/incomplete_program.mlir"],
            ParseError,
            "Could not parse entire input",
        ),
        (
            ["tests/xdsl_opt/incomplete_program_residual.mlir"],
            ParseError,
            "Could not parse entire input",
        ),
        (
            ["tests/xdsl_opt/incomplete_program.mlir"],
            ParseError,
            "Could not parse entire input",
        ),
        (
            ["tests/xdsl_opt/empty_program.wrong"],
            ValueError,
            "Unrecognized file extension 'wrong'",
        ),
        (
            ["tests/xdsl_opt/unverified_program.mlir"],
            DiagnosticException,
            "operand at position 0 does not verify",
        ),
    ],
)
def test_error_on_run(
    args: list[str], error_type: type[Exception], expected_error: str
):
    opt = xDSLOptMain(args=args)

    with pytest.raises(error_type, match=expected_error):
        opt.run()


@pytest.mark.parametrize(
    "args, error_type, expected_error",
    [
        (
            ["tests/xdsl_opt/empty_program.mlir", "-p", "wrong"],
            ValueError,
            "Unrecognized passes: ['wrong']",
        ),
    ],
)
def test_error_on_construction(
    args: list[str], error_type: type[Exception], expected_error: str
):
    with pytest.raises(error_type, match=re.escape(expected_error)):
        xDSLOptMain(args=args)


@pytest.mark.parametrize(
    "args, expected_error",
    [
        (
            ["tests/xdsl_opt/empty_program.mlir", "-t", "wrong"],
            "invalid choice: 'wrong'",
        ),
    ],
)
def test_error_on_argparse(
    capsys: pytest.CaptureFixture[str], args: list[str], expected_error: str
):
    with pytest.raises(SystemExit):
        xDSLOptMain(args=args)
    out, err = capsys.readouterr()
    assert out == ""
    assert expected_error in err


def test_print_to_file():
    filename_in = "tests/xdsl_opt/empty_program.mlir"
    filename_out = "tests/xdsl_opt/empty_program.out"

    opt = xDSLOptMain(args=[filename_in, "-o", filename_out])
    opt.run()

    with open(filename_in) as file:
        inp = file.read()
    with open(filename_out) as file:
        expected = file.read()

    assert inp.strip() == expected.strip()


def test_operation_deletion():
    filename_in = "tests/xdsl_opt/simple_program.mlir"
    filename_out = "tests/xdsl_opt/empty_program.mlir"

    class xDSLOptMainPass(xDSLOptMain):
        def register_all_passes(self):
            class RemoveConstantPass(ModulePass):
                name = "remove-constant"

                def apply(self, ctx: Context, op: builtin.ModuleOp):
                    if op.ops.first is not None:
                        op.ops.first.detach()

            self.register_pass("remove-constant", lambda: RemoveConstantPass)

    opt = xDSLOptMainPass(args=[filename_in, "-p", "remove-constant"])

    f = StringIO("")
    with redirect_stdout(f):
        opt.run()
    with open(filename_out) as file:
        expected = file.read()

    assert f.getvalue().strip() == expected.strip()


def test_print_between_passes():
    filename_in = "tests/xdsl_opt/empty_program.mlir"
    passes = ["shape-inference", "dce", "frontend-desymrefy"]
    flags = ["--print-between-passes", "-p", ",".join(passes)]

    f = StringIO("")

    opt = xDSLOptMain(args=[*flags, filename_in])

    with redirect_stdout(f):
        opt.run()

    output = f.getvalue()
    assert len([l for l in output.split("\n") if "builtin.module" in l]) == len(passes)


def test_verify_diagnostics_output():
    """
    Diagnostic exceptions raised when printing output should be redirected to stdout if
    --verify-diagnostics flag is provided.
    """

    class TestMain(xDSLOptMain):
        def register_all_targets(self):
            def _my_target(prog: builtin.ModuleOp, output: IO[str]):
                raise DiagnosticException("fail")

            self.available_targets["fail"] = _my_target

        def get_input_stream(self) -> tuple[IO[str], str]:
            fake_input = StringIO("builtin.module {}")
            return (fake_input, "mlir")

    opt = TestMain(args=["-t", "fail"])
    f = StringIO("")
    with redirect_stdout(f):
        with pytest.raises(DiagnosticException, match="fail"):
            opt.run()
    assert f.getvalue() == ""

    opt = TestMain(args=["--verify-diagnostics", "-t", "fail"])
    f = StringIO("")
    with redirect_stdout(f):
        opt.run()
    assert f.getvalue() == "fail\n"


def test_split_input():
    filename_in = "tests/xdsl_opt/empty_program.mlir"
    filename_out = "tests/xdsl_opt/split_input_file.out"
    flag = "--split-input-file"

    opt = xDSLOptMain(args=[filename_in, flag, "-o", filename_out])
    opt.run()
    with open(filename_in) as file:
        inp = file.read()
    with open(filename_out) as file:
        expected = file.read()

    assert inp.strip() == expected.strip()



#!/usr/bin/env python3
"""Benchmarks for the lexer of the xDSL implementation."""

from benchmarks.workloads import WorkloadBuilder
from xdsl.utils.lexer import Input
from xdsl.utils.mlir_lexer import MLIRLexer, MLIRTokenKind


class Lexer:
    """Benchmark the xDSL lexer on MLIR files."""

    WORKLOAD_EMPTY = WorkloadBuilder.empty()
    WORKLOAD_CONSTANT_100 = WorkloadBuilder.constant_folding(100)
    WORKLOAD_CONSTANT_1000 = WorkloadBuilder.constant_folding(1000)
    WORKLOAD_LARGE_DENSE_ATTR = WorkloadBuilder.large_dense_attr()
    WORKLOAD_LARGE_DENSE_ATTR_HEX = WorkloadBuilder.large_dense_attr_hex()

    def time_empty_program(self) -> None:
        """Time lexing an empty program."""
        lexer_input = Input(Lexer.WORKLOAD_EMPTY, "empty")
        lexer = MLIRLexer(lexer_input)
        while lexer.lex().kind is not MLIRTokenKind.EOF:
            pass

    def time_constant_100(self) -> None:
        """Time lexing constant folding for 100 items."""
        lexer_input = Input(Lexer.WORKLOAD_CONSTANT_100, "constant_100")
        lexer = MLIRLexer(lexer_input)
        while lexer.lex().kind is not MLIRTokenKind.EOF:
            pass

    def time_constant_1000(self) -> None:
        """Time lexing constant folding for 1000 items."""
        lexer_input = Input(Lexer.WORKLOAD_CONSTANT_1000, "constant_1000")
        lexer = MLIRLexer(lexer_input)
        while lexer.lex().kind is not MLIRTokenKind.EOF:
            pass

    def ignore_time_dense_attr(self) -> None:
        """Time lexing a 1024x1024xi8 dense attribute."""
        lexer_input = Input(Lexer.WORKLOAD_LARGE_DENSE_ATTR, "dense_attr")
        lexer = MLIRLexer(lexer_input)
        while lexer.lex().kind is not MLIRTokenKind.EOF:
            pass

    def ignore_time_dense_attr_hex(self) -> None:
        """Time lexing a 1024x1024xi8 dense attribute given as a hex string."""
        lexer_input = Input(Lexer.WORKLOAD_LARGE_DENSE_ATTR_HEX, "dense_attr_hex")
        lexer = MLIRLexer(lexer_input)
        while lexer.lex().kind is not MLIRTokenKind.EOF:
            pass


if __name__ == "__main__":
    from bench_utils import Benchmark, profile

    LEXER = Lexer()
    profile(
        {
            "Lexer.empty_program": Benchmark(LEXER.time_empty_program),
            "Lexer.constant_100": Benchmark(LEXER.time_constant_100),
            "Lexer.constant_1000": Benchmark(LEXER.time_constant_1000),
            "Lexer.dense_attr": Benchmark(LEXER.ignore_time_dense_attr),
            "Lexer.dense_attr_hex": Benchmark(LEXER.ignore_time_dense_attr_hex),
        }
    )



from __future__ import annotations

import contextlib
import threading
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass, field
from typing import TypeAlias, overload

from typing_extensions import TypeVar

from xdsl.dialects.builtin import ArrayAttr
from xdsl.ir import Attribute, Block, BlockArgument, Operation, OperationInvT, Region
from xdsl.rewriter import BlockInsertPoint, InsertPoint, Rewriter


@dataclass(eq=False)
class BuilderListener:
    """A listener for builder events."""

    operation_insertion_handler: list[Callable[[Operation], None]] = field(
        default_factory=list[Callable[[Operation], None]], kw_only=True
    )
    """Callbacks that are called when an operation is inserted by the builder."""

    block_creation_handler: list[Callable[[Block], None]] = field(
        default_factory=list[Callable[[Block], None]], kw_only=True
    )
    """Callback that are called when a block is created by the builder."""

    def handle_operation_insertion(self, op: Operation) -> None:
        """Pass the operation that was just inserted to callbacks."""
        for callback in self.operation_insertion_handler:
            callback(op)

    def handle_block_creation(self, block: Block) -> None:
        """Pass the block that was just created to callbacks."""
        for callback in self.block_creation_handler:
            callback(block)

    def extend_from_listener(self, listener: BuilderListener) -> None:
        """Forward all callbacks from `listener` to this listener."""
        self.operation_insertion_handler.extend(listener.operation_insertion_handler)
        self.block_creation_handler.extend(listener.block_creation_handler)


InsertOpInvT = TypeVar("InsertOpInvT", bound=Operation | Sequence[Operation])


@dataclass
class Builder(BuilderListener):
    """
    A helper class to construct IRs, by keeping track of where to insert an
    operation. It mimics the OpBuilder class from MLIR.

    See external [documentation](https://mlir.llvm.org/doxygen/classmlir_1_1OpBuilder.html).
    """

    insertion_point: InsertPoint
    """Operations will be inserted at this location."""

    def insert(self, op: OperationInvT) -> OperationInvT:
        """
        Inserts op at the current location and returns it.
        """
        return self.insert_op(op)

    def insert_op(
        self,
        op: InsertOpInvT,
        insertion_point: InsertPoint | None = None,
    ) -> InsertOpInvT:
        """Inserts op(s) at the current insertion point."""
        ops = (op,) if isinstance(op, Operation) else op
        if not ops:
            return ops

        implicit_builder = _current_builder.builder
        if implicit_builder is not None and implicit_builder is not self:
            raise ValueError(
                "Cannot insert operation explicitly when an implicit builder exists."
            )

        Rewriter.insert_op(
            op, self.insertion_point if insertion_point is None else insertion_point
        )

        for op_ in ops:
            self.handle_operation_insertion(op_)

        return op

    def create_block(
        self, insert_point: BlockInsertPoint, arg_types: Iterable[Attribute] = ()
    ) -> Block:
        """
        Create a block at the given location, and set the operation insertion point
        at the end of the inserted block.
        """
        block = Block(arg_types=arg_types)
        Rewriter.insert_block(block, insert_point)

        self.insertion_point = InsertPoint.at_end(block)

        self.handle_block_creation(block)
        return block

    @staticmethod
    def _region_no_args(func: Callable[[Builder], None]) -> Region:
        """
        Generates a single-block region.
        """
        block = Block()
        builder = Builder(InsertPoint.at_end(block))
        func(builder)
        return Region(block)

    @staticmethod
    def _region_args(
        input_types: Sequence[Attribute] | ArrayAttr[Attribute],
    ) -> Callable[[_CallableRegionFuncType], Region]:
        """
        Decorator for constructing a single-block region, containing the implementation of a
        region with some input arguments.
        """

        if isinstance(input_types, ArrayAttr):
            input_types = input_types.data

        def wrapper(func: _CallableRegionFuncType) -> Region:
            block = Block(arg_types=input_types)
            builder = Builder(InsertPoint.at_start(block))

            func(builder, block.args)

            region = Region(block)
            return region

        return wrapper

    @overload
    @staticmethod
    def region(
        input: Sequence[Attribute] | ArrayAttr[Attribute],
    ) -> Callable[[_CallableRegionFuncType], Region]:
        """
        Annotation used to construct a Region tuple from a function.
        The annotation can be used in two ways:

        For regions that have inputs or outputs:
        ```
        @Builder.region(input_types)
        def func(builder: Builder, args: tuple[BlockArgument, ...]) -> None:
            ...
        ```

        For regions that don't have inputs or outputs:
        ``` python
        @Builder.region
        def func(builder: Builder) -> None:
            ...
        ```
        """
        ...

    @overload
    @staticmethod
    def region(input: Callable[[Builder], None]) -> Region: ...

    @staticmethod
    def region(
        input: Sequence[Attribute] | ArrayAttr[Attribute] | Callable[[Builder], None],
    ) -> Callable[[_CallableRegionFuncType], Region] | Region:
        if isinstance(input, Callable):
            return Builder._region_no_args(input)
        else:
            return Builder._region_args(input)

    @staticmethod
    def _implicit_region_no_args(func: Callable[[], None]) -> Region:
        """
        Generates a single-block region.
        """
        block = Block()
        builder = Builder(InsertPoint.at_end(block))

        with ImplicitBuilder(builder):
            func()

        return Region(block)

    @staticmethod
    def _implicit_region_args(
        input_types: Sequence[Attribute] | ArrayAttr[Attribute],
    ) -> Callable[[_CallableImplicitRegionFuncType], Region]:
        """
        Decorator for constructing a single-block region, containing the implementation of a
        region with some input arguments.
        """

        if isinstance(input_types, ArrayAttr):
            input_types = input_types.data

        def wrapper(func: _CallableImplicitRegionFuncType) -> Region:
            block = Block(arg_types=input_types)
            builder = Builder(InsertPoint.at_end(block))

            with ImplicitBuilder(builder):
                func(block.args)

            region = Region(block)
            return region

        return wrapper

    @overload
    @staticmethod
    def implicit_region(
        input: Sequence[Attribute] | ArrayAttr[Attribute],
    ) -> Callable[[_CallableImplicitRegionFuncType], Region]:
        """
        Annotation used to construct a Region tuple from a function.
        The annotation can be used in two ways:

        For regions that have inputs or outputs:
        ```
        @Builder.implicit_region(input_types)
        def func(args: tuple[BlockArgument, ...]) -> None:
            ...
        ```

        For regions that don't have inputs or outputs:
        ``` python
        @Builder.implicit_region
        def func() -> None:
            ...
        ```
        """
        ...

    @overload
    @staticmethod
    def implicit_region(input: Callable[[], None]) -> Region: ...

    @staticmethod
    def implicit_region(
        input: Sequence[Attribute] | ArrayAttr[Attribute] | Callable[[], None],
    ) -> Callable[[_CallableImplicitRegionFuncType], Region] | Region:
        if isinstance(input, Callable):
            return Builder._implicit_region_no_args(input)
        else:
            return Builder._implicit_region_args(input)

    @staticmethod
    def assert_implicit():
        if _current_builder.builder is None:
            raise ValueError(
                "op_builder must be called within an implicit builder block"
            )


# Implicit builders


@dataclass
class _ThreadLocalBuilder(threading.local):
    """
    Stores the implicit builder for use in ImplicitBuilder, None by default.
    There is a builder per thread, guaranteed by inheriting from `threading.local`.
    """

    builder: Builder | None = None


_current_builder = _ThreadLocalBuilder()


@contextlib.contextmanager
def ImplicitBuilder(
    arg: Builder | Block | Region | None,
):
    """
    Context manager for managing the current implicit builder context, consisting of the stack of builders in
    the current thread, and the current builder.

    Operations created within a `with` block of an implicit builder will be added to it.
    If there are nested implicit builder blocks, the operation will be added to the
    innermost one. Operations cannot be added to multiple blocks, and any attempt to do so
    will result in an exception.

    Example:

    ``` python
    from xdsl.dialects import arith

    with ImplicitBuilder(block):
        arith.Constant(IntegerAttr(5, 32))

    assert len(block.ops) == 1
    assert isinstance(block.ops.first, arith.Constant)
    ```
    """
    match arg:
        case None:
            # None option added as convenience to allow for extending optional regions
            # in ops easily.
            raise ValueError("Cannot pass None to implicit_builder")
        case Region():
            builder = Builder(InsertPoint.at_end(arg.block))
        case Block():
            builder = Builder(InsertPoint.at_end(arg))
        case Builder():
            builder = arg

    old_builder = _current_builder.builder
    old_post_init = None
    if old_builder is None:
        old_post_init = _override_operation_post_init()

    _current_builder.builder = builder
    try:
        yield builder.insertion_point.block.args
    finally:
        _current_builder.builder = old_builder
        if old_builder is None:
            Operation.__post_init__ = old_post_init  # pyright: ignore[reportAttributeAccessIssue]


_CallableRegionFuncType: TypeAlias = Callable[
    [Builder, tuple[BlockArgument, ...]], None
]
_CallableImplicitRegionFuncType: TypeAlias = Callable[[tuple[BlockArgument, ...]], None]


def _op_init_callback(op: Operation):
    if (b := _current_builder.builder) is not None:
        b.insert(op)


def _override_operation_post_init() -> Callable[[Operation], None]:
    old_post_init = Operation.__post_init__

    def new_post_init(self: Operation) -> None:
        old_post_init(self)
        _op_init_callback(self)

    Operation.__post_init__ = new_post_init
    return old_post_init



import re
from collections.abc import Sequence
from typing import ClassVar

import pytest

from xdsl.backend.block_naive_allocator import BlockNaiveAllocator
from xdsl.backend.register_allocatable import (
    HasRegisterConstraints,
    RegisterAllocatableOperation,
    RegisterConstraints,
)
from xdsl.backend.register_allocator import ValueAllocator
from xdsl.backend.register_stack import OutOfRegisters, RegisterStack
from xdsl.backend.register_type import RegisterType
from xdsl.builder import Builder
from xdsl.ir import Attribute, Block, SSAValue
from xdsl.irdl import (
    AttrSizedOperandSegments,
    AttrSizedResultSegments,
    IRDLOperation,
    VarConstraint,
    base,
    irdl_attr_definition,
    irdl_op_definition,
    operand_def,
    result_def,
    var_operand_def,
    var_result_def,
)
from xdsl.utils.exceptions import DiagnosticException
from xdsl.utils.test_value import create_ssa_value


@irdl_attr_definition
class TestRegister(RegisterType):
    name = "test.reg"

    @classmethod
    def index_by_name(cls) -> dict[str, int]:
        return {"x0": 0, "x1": 1, "a0": 0, "a1": 1}

    @classmethod
    def infinite_register_prefix(cls):
        return "y"


@irdl_op_definition
class TestAllocatableOp(IRDLOperation, HasRegisterConstraints):
    name = "test.allocatable"

    in_operands = var_operand_def()
    inout_operands = var_operand_def()
    out_results = var_result_def()
    inout_results = var_result_def()

    irdl_options = [AttrSizedOperandSegments(), AttrSizedResultSegments()]

    def __init__(
        self,
        in_operands: Sequence[SSAValue],
        inout_operands: Sequence[SSAValue],
        out_result_types: Sequence[Attribute],
        inout_result_types: Sequence[Attribute],
    ):
        super().__init__(
            operands=(in_operands, inout_operands),
            result_types=(out_result_types, inout_result_types),
        )

    def get_register_constraints(self) -> RegisterConstraints:
        return RegisterConstraints(
            self.in_operands,
            self.out_results,
            tuple(zip(self.inout_operands, self.inout_results)),
        )


def op(
    ins: Sequence[SSAValue],
    *out_result_types: Attribute,
    inouts: Sequence[SSAValue] = (),
):
    return TestAllocatableOp(
        ins, inouts, out_result_types, tuple(val.type for val in inouts)
    )


def test_gather_allocated():
    u = TestRegister.unallocated()
    x0 = TestRegister.from_name("x0")
    x1 = TestRegister.from_name("x0")

    @Builder.implicit_region
    def no_preallocated_body() -> None:
        (v1,) = op((), u).results
        (v2,) = op((), u).results
        op((v1, v2), u)

    pa_regs = set(
        RegisterAllocatableOperation.iter_all_used_registers(no_preallocated_body)
    )

    assert pa_regs == set()

    @Builder.implicit_region
    def one_preallocated_body() -> None:
        (v1,) = op((), u).results
        (v2,) = op((), x0).results
        op((v1, v2), u)

    pa_regs = set(
        RegisterAllocatableOperation.iter_all_used_registers(one_preallocated_body)
    )

    assert pa_regs == {x0}

    @Builder.implicit_region
    def repeated_preallocated_body() -> None:
        (v1,) = op((), u).results
        (v2,) = op((), x0).results
        (v3,) = op((), x0).results
        op((v1, v2, v3), u)

    pa_regs = set(
        RegisterAllocatableOperation.iter_all_used_registers(repeated_preallocated_body)
    )

    assert pa_regs == {x0}

    @Builder.implicit_region
    def multiple_preallocated_body() -> None:
        (v1,) = op((), u).results
        (v2,) = op((), x0).results
        (v3,) = op((), x1).results
        op((v1, v2, v3), u)

    pa_regs = set(
        RegisterAllocatableOperation.iter_all_used_registers(multiple_preallocated_body)
    )

    assert pa_regs == {x0, x1}


def test_new_type_for_value():
    @irdl_attr_definition
    class OtherRegister(RegisterType):
        name = "test.other_reg"

        @classmethod
        def index_by_name(cls) -> dict[str, int]:
            return {"x0": 0, "x1": 1, "a0": 0, "a1": 1}

        @classmethod
        def infinite_register_prefix(cls):
            return "y"

    u = TestRegister.unallocated()
    a0 = TestRegister.from_name("a0")
    a1 = TestRegister.from_name("a1")

    available_registers = RegisterStack.get((a0, a1))
    allocator = ValueAllocator(available_registers, TestRegister)

    r0, r1, r2, r3 = op(
        (), u, a0, OtherRegister.unallocated(), OtherRegister.from_index(1)
    ).results

    assert allocator.new_type_for_value(r0) == a1
    assert allocator.new_type_for_value(r1) is None
    assert allocator.new_type_for_value(r2) is None
    assert allocator.new_type_for_value(r3) is None


def test_allocate_value():
    u = TestRegister.unallocated()
    a0 = TestRegister.from_name("a0")
    a1 = TestRegister.from_name("a1")
    y0 = TestRegister.from_name("y0")

    register_stack = RegisterStack.get((a0, a1), allow_infinite=True)
    allocator = ValueAllocator(register_stack, TestRegister)

    op0 = op((), u, u, u, u)
    r00, r01, r02, _r03 = op0.results
    op1 = op((r00, r01), u, u, inouts=(r02,))
    results00 = tuple(op0.results)
    results10 = tuple(op1.results)

    # Initial state
    assert register_stack.available_registers == {"test.reg": [0, 1]}
    assert allocator.new_value_by_old_value == {}
    assert op0.result_types == (u, u, u, u)
    assert op1.result_types == (u, u, u)

    # Allocate first result
    new_value = allocator.allocate_value(op1.out_results[0])
    assert new_value is not None

    results11 = tuple(op1.results)

    assert register_stack.available_registers == {"test.reg": [0]}
    assert allocator.new_value_by_old_value == {results10[0]: results11[0]}
    assert op0.result_types == (u, u, u, u)
    assert op1.result_types == (a1, u, u)

    # Allocate first result again (old value) -> no change of state
    new_value = allocator.allocate_value(results10[0])
    assert new_value is None
    assert register_stack.available_registers == {"test.reg": [0]}
    assert allocator.new_value_by_old_value == {results10[0]: results11[0]}
    assert op0.result_types == (u, u, u, u)
    assert op1.result_types == (a1, u, u)

    # Allocate first result again (new value) -> no change of state
    new_value = allocator.allocate_value(op1.out_results[0])
    assert new_value is None
    assert register_stack.available_registers == {"test.reg": [0]}
    assert allocator.new_value_by_old_value == {results10[0]: results11[0]}
    assert op0.result_types == (u, u, u, u)
    assert op1.result_types == (a1, u, u)

    # Allocate second result
    new_value = allocator.allocate_value(op1.out_results[1])
    assert new_value is not None

    results12 = tuple(op1.results)

    assert register_stack.available_registers == {"test.reg": []}
    assert allocator.new_value_by_old_value == {
        results10[0]: results11[0],
        results11[1]: results12[1],
    }
    assert op0.result_types == (u, u, u, u)
    assert op1.result_types == (a1, a0, u)

    # Allocate inout result and operands at the same time
    new_value = allocator.allocate_values_same_reg(
        (op1.inout_operands[0], op1.inout_results[0])
    )
    results03 = tuple(op0.results)
    results13 = tuple(op1.results)

    assert register_stack.available_registers == {"test.reg": []}
    assert allocator.new_value_by_old_value == {
        results10[0]: results11[0],
        results11[1]: results12[1],
        results12[2]: results13[2],
        results00[2]: results03[2],
    }
    assert op0.result_types == (u, u, y0, u)
    assert op1.result_types == (a1, a0, y0)

    # Free the allocated result values
    allocator.free_value(results13[0])
    allocator.free_value(results13[1])

    assert register_stack.available_registers == {"test.reg": [1, 0]}
    assert allocator.new_value_by_old_value == {
        results10[0]: results11[0],
        results11[1]: results12[1],
        results12[2]: results13[2],
        results00[2]: results03[2],
    }
    assert op0.result_types == (u, u, y0, u)
    assert op1.result_types == (a1, a0, y0)

    # Allocate operand
    allocator.allocate_value(op1.operands[0])
    results04 = tuple(op0.results)

    assert register_stack.available_registers == {"test.reg": [1]}
    assert allocator.new_value_by_old_value == {
        results10[0]: results11[0],
        results11[1]: results12[1],
        results12[2]: results13[2],
        results00[2]: results03[2],
        results00[0]: results04[0],
    }
    assert op0.result_types == (a0, u, y0, u)
    assert op1.result_types == (a1, a0, y0)


def test_allocate_values_same_reg():
    u = TestRegister.unallocated()
    a0 = TestRegister.from_name("a0")
    a1 = TestRegister.from_name("a1")
    y0 = TestRegister.from_name("y0")

    register_stack = RegisterStack.get((a0, a1))
    allocator = ValueAllocator(register_stack, TestRegister)

    # Empty
    assert not allocator.allocate_values_same_reg(())

    op0 = op((), u, u, u, u, u)

    # 1 unallocated
    assert allocator.allocate_values_same_reg((op0.results[0],))
    assert op0.result_types == (a1, u, u, u, u)

    # 1 allocated
    assert not allocator.allocate_values_same_reg((op0.results[0],))
    assert op0.result_types == (a1, u, u, u, u)

    # 2 unallocated
    assert allocator.allocate_values_same_reg((op0.results[1], op0.results[2]))
    assert op0.result_types == (a1, a0, a0, u, u)

    # 1 allocated 1 unallocated
    assert allocator.allocate_values_same_reg((op0.results[0], op0.results[3]))
    assert op0.result_types == (a1, a0, a0, a1, u)

    # 1 unallocated 1 allocated
    assert allocator.allocate_values_same_reg((op0.results[4], op0.results[0]))
    assert op0.result_types == (a1, a0, a0, a1, a1)

    with pytest.raises(
        DiagnosticException,
        match=re.escape(
            "Cannot allocate registers to the same register ['!test.reg<a0>', '!test.reg<a1>']"
        ),
    ):
        allocator.allocate_values_same_reg((op0.results[0], op0.results[1]))

    op1 = op((), y0)

    with pytest.raises(
        DiagnosticException,
        match=re.escape(
            "Cannot allocate registers to the same register ['!test.reg<a0>', '!test.reg<a1>', '!test.reg<y0>']"
        ),
    ):
        allocator.allocate_values_same_reg(
            (op0.results[0], op0.results[1], op1.results[0])
        )


def test_fail_error_message():
    @irdl_op_definition
    class InoutOp(HasRegisterConstraints, IRDLOperation):
        name = "test.inout"

        T: ClassVar = VarConstraint("T", base(TestRegister))

        a = operand_def(T)
        b = result_def(T)

        def get_register_constraints(self) -> RegisterConstraints:
            return RegisterConstraints((), (), ((self.a, self.b),))

    a = create_ssa_value(TestRegister.from_index(0))
    block = Block((InoutOp(operands=[a], result_types=[TestRegister.from_index(1)]),))

    allocator = BlockNaiveAllocator(RegisterStack.get(), TestRegister)

    with pytest.raises(
        DiagnosticException,
        match=re.escape(
            "Cannot allocate registers to the same register ['!test.reg<a0>', '!test.reg<a1>']"
        ),
    ) as e:
        allocator.allocate_block(block)

    assert getattr(e.value, "__notes__") == [
        """
^0:
  %0 = "test.inout"(%1) : (!test.reg<a0>) -> !test.reg<a1>
  ^^^^^^^^^^^^^^^^^----
  | Error allocating op
  ---------------------
"""
    ]


def test_out_of_registers():
    register_stack = RegisterStack()
    with pytest.raises(OutOfRegisters, match="Out of registers."):
        register_stack.pop(TestRegister)



from io import StringIO

from xdsl.backend.wgsl.wgsl_printer import WGSLPrinter
from xdsl.dialects import arith, gpu, memref, test
from xdsl.dialects.builtin import IndexType, IntegerAttr, IntegerType, i32
from xdsl.utils.test_value import create_ssa_value

lhs_op = test.TestOp(result_types=[IndexType()])
rhs_op = test.TestOp(result_types=[IndexType()])


def test_gpu_global_id():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    global_id_x = gpu.GlobalIdOp(gpu.DimensionAttr(gpu.DimensionEnum.X))
    printer.print(global_id_x)

    assert "let v0: u32 = global_invocation_id.x;" in file.getvalue()


def test_gpu_thread_id():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    thread_id_x = gpu.ThreadIdOp(gpu.DimensionAttr(gpu.DimensionEnum.X))
    printer.print(thread_id_x)

    assert "let v0: u32 = local_invocation_id.x;" in file.getvalue()


def test_gpu_block_id():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    block_id_x = gpu.BlockIdOp(gpu.DimensionAttr(gpu.DimensionEnum.X))
    printer.print(block_id_x)

    assert "let v0: u32 = workgroup_id.x;" in file.getvalue()


def test_gpu_grid_dim():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    num_workgroups = gpu.GridDimOp(gpu.DimensionAttr(gpu.DimensionEnum.X))
    printer.print(num_workgroups)

    assert "let v0: u32 = num_workgroups.x;" in file.getvalue()


def test_arith_constant_unsigned():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    cst = arith.ConstantOp(IntegerAttr(42, IndexType()))
    printer.print(cst)

    assert "let v0 : u32 = 42u;" in file.getvalue()


def test_arith_constant_unsigned_neg():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    cst = arith.ConstantOp(IntegerAttr(-1, IndexType()))
    cst.result.name_hint = "temp"
    printer.print(cst)

    assert "let vtemp : u32 = 4294967295u;" in file.getvalue()


def test_arith_constant_signed():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    cst = arith.ConstantOp(IntegerAttr(42, IntegerType(32)))
    cst.result.name_hint = "temp"
    printer.print(cst)

    assert "let vtemp : i32 = 42;" in file.getvalue()


def test_arith_addi():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    addi = arith.AddiOp(lhs_op, rhs_op)
    printer.print(addi)

    assert "let v0 = v1 + v2;" in file.getvalue()


def test_arith_subi():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    subi = arith.SubiOp(lhs_op, rhs_op)
    printer.print(subi)

    assert "let v0 = v1 - v2;" in file.getvalue()


def test_arith_muli():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    muli = arith.MuliOp(lhs_op, rhs_op)
    printer.print(muli)

    assert "let v0 = v1 * v2;" in file.getvalue()


def test_arith_addf():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    addf = arith.AddfOp(lhs_op, rhs_op)
    printer.print(addf)

    assert "let v0 = v1 + v2;" in file.getvalue()


def test_arith_subf():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    subf = arith.SubfOp(lhs_op, rhs_op)
    printer.print(subf)

    assert "let v0 = v1 - v2;" in file.getvalue()


def test_arith_mulf():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    mulf = arith.MulfOp(lhs_op, rhs_op)
    printer.print(mulf)

    assert "let v0 = v1 * v2;" in file.getvalue()


def test_memref_load():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    memref_type = memref.MemRefType(i32, [10, 10])
    memref_val = create_ssa_value(memref_type)
    load = memref.LoadOp.get(memref_val, [lhs_op.res[0], rhs_op.res[0]])
    printer.print(load)

    assert "let v1 = v0[10u * v1 + 1u * v2];" in file.getvalue()


def test_memref_store():
    file = StringIO("")
    printer = WGSLPrinter(stream=file)

    memref_type = memref.MemRefType(i32, [10, 10])
    memref_val = create_ssa_value(memref_type)
    load = memref.LoadOp.get(memref_val, [lhs_op.res[0], rhs_op.res[0]])
    store = memref.StoreOp.get(load.res, memref_val, [lhs_op.res[0], rhs_op.res[0]])
    printer.print(store)

    assert "v1[10u * v1 + 1u * v2] = v0;" in file.getvalue()



from __future__ import annotations

import abc
from collections.abc import Iterable, Sequence
from typing import ClassVar, cast

from typing_extensions import Self

from xdsl.dialects.builtin import (
    I64,
    AnyFloatConstr,
    ArrayAttr,
    BoolAttr,
    DenseArrayBase,
    DenseIntOrFPElementsAttr,
    IndexType,
    IntAttr,
    IntegerAttr,
    IntegerType,
    MemRefLayoutAttr,
    MemRefType,
    NoneAttr,
    SignlessIntegerConstraint,
    StridedLayoutAttr,
    StringAttr,
    SymbolNameConstraint,
    SymbolRefAttr,
    UnitAttr,
    UnrankedMemRefType,
    i32,
    i64,
)
from xdsl.dialects.utils import (
    split_dynamic_index_list,
)
from xdsl.dialects.utils.dynamic_index_list import (
    DynamicIndexList,
    verify_dynamic_index_list,
)
from xdsl.dialects.utils.reshape_ops_utils import (
    ContiguousArrayOfIntArray,
)
from xdsl.ir import Attribute, Dialect, Operation, SSAValue
from xdsl.irdl import (
    AnyAttr,
    AttrSizedOperandSegments,
    IRDLOperation,
    ParsePropInAttrDict,
    SameVariadicResultSize,
    VarConstraint,
    base,
    irdl_op_definition,
    lazy_traits_def,
    operand_def,
    opt_prop_def,
    prop_def,
    region_def,
    result_def,
    traits_def,
    var_operand_def,
    var_result_def,
)
from xdsl.parser import Parser
from xdsl.pattern_rewriter import RewritePattern
from xdsl.printer import Printer
from xdsl.traits import (
    HasCanonicalizationPatternsTrait,
    HasParent,
    IsTerminator,
    MemoryAllocEffect,
    MemoryFreeEffect,
    MemoryReadEffect,
    MemoryWriteEffect,
    NoMemoryEffect,
    SymbolOpInterface,
)
from xdsl.utils.bitwise_casts import is_power_of_two
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.hints import isa


@irdl_op_definition
class LoadOp(IRDLOperation):
    name = "memref.load"

    T: ClassVar = VarConstraint("T", AnyAttr())

    nontemporal = opt_prop_def(BoolAttr, default_value=BoolAttr.from_bool(False))

    memref = operand_def(MemRefType.constr(T))
    indices = var_operand_def(IndexType())
    res = result_def(T)

    traits = traits_def(MemoryReadEffect())

    irdl_options = [ParsePropInAttrDict()]
    assembly_format = "$memref `[` $indices `]` attr-dict `:` type($memref)"

    # TODO varargs for indexing, which must match the memref dimensions
    # Problem: memref dimensions require variadic type parameters,
    # which is subject to change

    def verify_(self):
        memref_type = self.memref.type
        if not isinstance(memref_type, MemRefType):
            raise VerifyException("expected a memreftype")

        memref_type = cast(MemRefType, memref_type)

        if memref_type.get_num_dims() != len(self.indices):
            raise Exception("expected an index for each dimension")

    @classmethod
    def get(
        cls, ref: SSAValue | Operation, indices: Sequence[SSAValue | Operation]
    ) -> Self:
        ssa_value = SSAValue.get(ref, type=MemRefType)
        return cls(operands=[ref, indices], result_types=[ssa_value.type.element_type])


@irdl_op_definition
class StoreOp(IRDLOperation):
    T: ClassVar = VarConstraint("T", AnyAttr())

    name = "memref.store"

    nontemporal = opt_prop_def(BoolAttr, default_value=BoolAttr.from_bool(False))

    value = operand_def(T)
    memref = operand_def(MemRefType.constr(T))
    indices = var_operand_def(IndexType())

    traits = traits_def(MemoryWriteEffect())

    irdl_options = [ParsePropInAttrDict()]
    assembly_format = "$value `,` $memref `[` $indices `]` attr-dict `:` type($memref)"

    def verify_(self):
        if not isinstance(memref_type := self.memref.type, MemRefType):
            raise VerifyException("expected a memreftype")

        memref_type = cast(MemRefType, memref_type)

        if memref_type.get_num_dims() != len(self.indices):
            raise Exception("Expected an index for each dimension")

    @classmethod
    def get(
        cls,
        value: Operation | SSAValue,
        ref: Operation | SSAValue,
        indices: Sequence[Operation | SSAValue],
    ) -> Self:
        return cls(operands=[value, ref, indices])


class AllocOpHasCanonicalizationPatterns(HasCanonicalizationPatternsTrait):
    @classmethod
    def get_canonicalization_patterns(cls) -> tuple[RewritePattern, ...]:
        from xdsl.transforms.canonicalization_patterns.memref import ElideUnusedAlloc

        return (ElideUnusedAlloc(),)


@irdl_op_definition
class AllocOp(IRDLOperation):
    name = "memref.alloc"

    dynamic_sizes = var_operand_def(IndexType)
    symbol_operands = var_operand_def(IndexType)

    memref = result_def(MemRefType)

    # TODO how to constraint the IntegerAttr type?
    alignment = opt_prop_def(IntegerAttr)

    irdl_options = [AttrSizedOperandSegments(as_property=True)]

    traits = traits_def(AllocOpHasCanonicalizationPatterns(), MemoryAllocEffect())

    def __init__(
        self,
        dynamic_sizes: Sequence[SSAValue],
        symbol_operands: Sequence[SSAValue],
        result_type: Attribute,
        alignment: Attribute | None = None,
    ):
        super().__init__(
            operands=(dynamic_sizes, symbol_operands),
            result_types=(result_type,),
            properties={"alignment": alignment},
        )

    @classmethod
    def get(
        cls,
        return_type: Attribute,
        alignment: int | IntegerAttr | None = None,
        shape: Iterable[int | IntAttr] | None = None,
        dynamic_sizes: Sequence[SSAValue | Operation] | None = None,
        layout: MemRefLayoutAttr | NoneAttr = NoneAttr(),
        memory_space: Attribute = NoneAttr(),
    ) -> Self:
        if shape is None:
            shape = [1]

        if dynamic_sizes is None:
            dynamic_sizes = []

        if isinstance(alignment, int):
            alignment = IntegerAttr.from_int_and_width(alignment, 64)

        return cls(
            tuple(SSAValue.get(ds) for ds in dynamic_sizes),
            (),
            MemRefType(return_type, shape, layout, memory_space),
            alignment,
        )

    def verify_(self) -> None:
        memref_type = self.memref.type

        dyn_dims = [x for x in memref_type.shape.data if x.data == -1]
        if len(dyn_dims) != len(self.dynamic_sizes):
            raise VerifyException(
                "op dimension operand count does not equal memref dynamic dimension count."
            )

    def print(self, printer: Printer):
        printer.print_string("(")
        printer.print_list(self.dynamic_sizes, printer.print_ssa_value)
        printer.print_string(")")
        if self.symbol_operands:
            printer.print_string("[")
            printer.print_list(self.symbol_operands, printer.print_ssa_value)
            printer.print_string("]")

        printer.print_op_attributes(
            self.properties | self.attributes,
            print_keyword=False,
            reserved_attr_names="operandSegmentSizes",
        )

        printer.print_string(" : ")
        printer.print_attribute(self.memref.type)

    @classmethod
    def parse(cls, parser: Parser) -> Self:
        #  %alloc = memref.alloc(%a)[%s] {alignment = 64 : i64} : memref<3x2xf32>

        unresolved_dynamic_sizes = parser.parse_comma_separated_list(
            parser.Delimiter.PAREN, parser.parse_unresolved_operand
        )
        unresolved_symbol_operands = parser.parse_optional_comma_separated_list(
            parser.Delimiter.SQUARE, parser.parse_unresolved_operand
        )
        if unresolved_symbol_operands is None:
            unresolved_symbol_operands = []

        attrs = parser.parse_optional_attr_dict()

        parser.parse_punctuation(":")
        res_type = parser.parse_attribute()

        index = IndexType()
        dynamic_sizes = tuple(
            parser.resolve_operand(uop, index) for uop in unresolved_dynamic_sizes
        )
        symbol_operands = tuple(
            parser.resolve_operand(uop, index) for uop in unresolved_symbol_operands
        )

        if "alignment" in attrs:
            alignment = attrs["alignment"]
            del attrs["alignment"]
        else:
            alignment = None

        op = cls(
            dynamic_sizes,
            symbol_operands,
            res_type,
            alignment,
        )

        op.attributes |= attrs

        return op


@irdl_op_definition
class AllocaScopeOp(IRDLOperation):
    name = "memref.alloca_scope"

    res = var_result_def()

    scope = region_def()


@irdl_op_definition
class AllocaScopeReturnOp(IRDLOperation):
    name = "memref.alloca_scope.return"

    ops = var_operand_def()

    traits = traits_def(IsTerminator(), HasParent(AllocaScopeOp))

    def verify_(self) -> None:
        parent = cast(AllocaScopeOp, self.parent_op())
        if self.ops.types != parent.result_types:
            raise VerifyException(
                "Expected operand types to match parent's return types."
            )


@irdl_op_definition
class AllocaOp(IRDLOperation):
    name = "memref.alloca"

    dynamic_sizes = var_operand_def(IndexType)
    symbol_operands = var_operand_def(IndexType)

    memref = result_def(MemRefType)

    # TODO how to constraint the IntegerAttr type?
    alignment = opt_prop_def(IntegerAttr)

    traits = traits_def(MemoryAllocEffect())

    irdl_options = [AttrSizedOperandSegments(as_property=True)]

    @staticmethod
    def get(
        return_type: Attribute,
        alignment: int | IntegerAttr | None = None,
        shape: Iterable[int | IntAttr] | None = None,
        dynamic_sizes: Sequence[SSAValue | Operation] | None = None,
        layout: MemRefLayoutAttr | NoneAttr = NoneAttr(),
        memory_space: Attribute = NoneAttr(),
    ) -> AllocaOp:
        if shape is None:
            shape = [1]

        if dynamic_sizes is None:
            dynamic_sizes = []

        if isinstance(alignment, int):
            alignment = IntegerAttr.from_int_and_width(alignment, 64)

        return AllocaOp.build(
            operands=[dynamic_sizes, []],
            result_types=[MemRefType(return_type, shape, layout, memory_space)],
            properties={
                "alignment": alignment,
            },
        )

    def verify_(self) -> None:
        memref_type = self.memref.type

        dyn_dims = [x for x in memref_type.shape.data if x.data == -1]
        if len(dyn_dims) != len(self.dynamic_sizes):
            raise VerifyException(
                "op dimension operand count does not equal memref dynamic dimension count."
            )


@irdl_op_definition
class AtomicRMWOp(IRDLOperation):
    name = "memref.atomic_rmw"

    T: ClassVar = VarConstraint("T", AnyFloatConstr | SignlessIntegerConstraint)

    value = operand_def(T)
    memref = operand_def(MemRefType.constr(T))
    indices = var_operand_def(IndexType)

    kind = prop_def(IntegerAttr[I64])

    result = result_def(T)

    traits = traits_def(MemoryWriteEffect(), MemoryReadEffect())


@irdl_op_definition
class DeallocOp(IRDLOperation):
    name = "memref.dealloc"
    memref = operand_def(base(MemRefType) | base(UnrankedMemRefType))

    traits = traits_def(MemoryFreeEffect())

    @staticmethod
    def get(operand: Operation | SSAValue) -> DeallocOp:
        return DeallocOp.build(operands=[operand])

    assembly_format = "$memref attr-dict `:` type($memref)"


@irdl_op_definition
class GetGlobalOp(IRDLOperation):
    name = "memref.get_global"
    memref = result_def(MemRefType)
    name_ = prop_def(SymbolRefAttr, prop_name="name")

    traits = traits_def(NoMemoryEffect())

    assembly_format = "$name `:` type($memref) attr-dict"

    def __init__(self, name: str | SymbolRefAttr, return_type: Attribute):
        if isinstance(name, str):
            name = SymbolRefAttr(name)
        super().__init__(result_types=[return_type], properties={"name": name})

    # TODO how to verify the types, as the global might be defined in another
    # compilation unit


@irdl_op_definition
class GlobalOp(IRDLOperation):
    name = "memref.global"

    sym_name = prop_def(SymbolNameConstraint())
    sym_visibility = prop_def(StringAttr)
    type = prop_def(MemRefType)
    initial_value = prop_def(UnitAttr | DenseIntOrFPElementsAttr)
    constant = opt_prop_def(UnitAttr)
    alignment = opt_prop_def(IntegerAttr[I64])

    traits = traits_def(SymbolOpInterface(), MemoryAllocEffect())

    def verify_(self) -> None:
        if self.alignment is not None:
            alignment_value = self.alignment.value.data
            # Alignment has to be a power of two
            if not (is_power_of_two(alignment_value)):
                raise VerifyException(
                    f"Alignment attribute {alignment_value} is not a power of 2"
                )

    @staticmethod
    def get(
        sym_name: StringAttr,
        sym_type: Attribute,
        initial_value: Attribute,
        sym_visibility: StringAttr = StringAttr("private"),
        constant: UnitAttr | None = None,
        alignment: int | IntegerAttr[IntegerType] | None = None,
    ) -> GlobalOp:
        if isinstance(alignment, int):
            alignment = IntegerAttr.from_int_and_width(alignment, 64)

        return GlobalOp.build(
            properties={
                "sym_name": sym_name,
                "type": sym_type,
                "initial_value": initial_value,
                "sym_visibility": sym_visibility,
                "constant": constant,
                "alignment": alignment,
            }
        )


@irdl_op_definition
class DimOp(IRDLOperation):
    name = "memref.dim"

    source = operand_def(base(MemRefType) | base(UnrankedMemRefType))
    index = operand_def(IndexType)

    result = result_def(IndexType)

    traits = traits_def(NoMemoryEffect())

    @staticmethod
    def from_source_and_index(
        source: SSAValue | Operation, index: SSAValue | Operation
    ):
        return DimOp.build(operands=[source, index], result_types=[IndexType()])


@irdl_op_definition
class RankOp(IRDLOperation):
    name = "memref.rank"

    source = operand_def(MemRefType)

    rank = result_def(IndexType)

    traits = traits_def(NoMemoryEffect())

    @staticmethod
    def from_memref(memref: Operation | SSAValue):
        return RankOp.build(operands=[memref], result_types=[IndexType()])


class AlterShapeOperation(IRDLOperation, abc.ABC):
    result = result_def(MemRefType)
    reassociation = prop_def(ContiguousArrayOfIntArray())

    traits = traits_def(NoMemoryEffect())


@irdl_op_definition
class CollapseShapeOp(AlterShapeOperation):
    """
    https://mlir.llvm.org/docs/Dialects/MemRef/#memrefcollapse_shape-memrefcollapseshapeop
    """

    name = "memref.collapse_shape"

    src = operand_def(MemRefType)

    assembly_format = (
        "$src $reassociation attr-dict `:` type($src) `into` type($result)"
    )


@irdl_op_definition
class ExpandShapeOp(AlterShapeOperation):
    """
    https://mlir.llvm.org/docs/Dialects/MemRef/#memrefexpand_shape-memrefexpandshapeop
    """

    name = "memref.expand_shape"

    src = operand_def(MemRefType)
    output_shape = var_operand_def(IndexType)

    static_output_shape = prop_def(DenseArrayBase.constr(i64))

    assembly_format = (
        "$src $reassociation `output_shape`"
        "custom<DynamicIndexList>($output_shape, $static_output_shape) attr-dict `:`"
        "type($src) `into` type($result)"
    )

    custom_directives = (DynamicIndexList,)


@irdl_op_definition
class ExtractStridedMetaDataOp(IRDLOperation):
    """
    https://mlir.llvm.org/docs/Dialects/MemRef/#memrefextract_strided_metadata-memrefextractstridedmetadataop
    """

    name = "memref.extract_strided_metadata"

    source = operand_def(MemRefType)

    base_buffer = result_def(MemRefType)
    offset = result_def(IndexType)
    sizes = var_result_def(IndexType)
    strides = var_result_def(IndexType)

    traits = traits_def(NoMemoryEffect())

    irdl_options = [SameVariadicResultSize()]

    def __init__(self, source: SSAValue | Operation):
        """
        Create an ExtractStridedMetaDataOp that extracts the metadata from the
        operation (source) that produces a memref.
        """
        source_type = SSAValue.get(source, type=MemRefType).type
        source_shape = source_type.get_shape()
        # Return a rank zero memref with the memref type
        base_buffer_type = MemRefType(
            source_type.element_type,
            [],
            NoneAttr(),
            source_type.memory_space,
        )
        offset_type = IndexType()
        # There are as many strides/sizes as there are shape dimensions
        strides_type = [IndexType()] * len(source_shape)
        sizes_type = [IndexType()] * len(source_shape)
        return_type = [base_buffer_type, offset_type, strides_type, sizes_type]
        super().__init__(operands=[source], result_types=return_type)


@irdl_op_definition
class ExtractAlignedPointerAsIndexOp(IRDLOperation):
    name = "memref.extract_aligned_pointer_as_index"

    source = operand_def(MemRefType)

    aligned_pointer = result_def(IndexType)

    traits = traits_def(NoMemoryEffect())

    @staticmethod
    def get(source: SSAValue | Operation):
        return ExtractAlignedPointerAsIndexOp.build(
            operands=[source], result_types=[IndexType()]
        )


class MemRefHasCanonicalizationPatternsTrait(HasCanonicalizationPatternsTrait):
    @classmethod
    def get_canonicalization_patterns(cls) -> tuple[RewritePattern, ...]:
        from xdsl.transforms.canonicalization_patterns.memref import (
            MemRefSubviewOfSubviewFolding,
        )

        return (MemRefSubviewOfSubviewFolding(),)


@irdl_op_definition
class SubviewOp(IRDLOperation):
    DYNAMIC_INDEX: ClassVar[int] = -9223372036854775808
    """
    Constant value used to denote dynamic indices in offsets, sizes, and strides.
    Same constant as in MLIR.
    """

    name = "memref.subview"

    source = operand_def(MemRefType)
    offsets = var_operand_def(IndexType)
    sizes = var_operand_def(IndexType)
    strides = var_operand_def(IndexType)
    static_offsets = prop_def(DenseArrayBase.constr(i64))
    static_sizes = prop_def(DenseArrayBase.constr(i64))
    static_strides = prop_def(DenseArrayBase.constr(i64))
    result = result_def(MemRefType)

    irdl_options = [AttrSizedOperandSegments(as_property=True)]

    traits = lazy_traits_def(
        lambda: (MemRefHasCanonicalizationPatternsTrait(), NoMemoryEffect())
    )

    assembly_format = (
        "$source ``"
        "custom<DynamicIndexList>($offsets, $static_offsets)"
        "custom<DynamicIndexList>($sizes, $static_sizes)"
        "custom<DynamicIndexList>($strides, $static_strides)"
        "attr-dict `:` type($source) `to` type($result)"
    )

    custom_directives = (DynamicIndexList,)

    def verify_(self) -> None:
        static_offsets = self.static_offsets.get_values()
        static_sizes = self.static_sizes.get_values()
        static_strides = self.static_strides.get_values()
        verify_dynamic_index_list(
            static_sizes, self.sizes, self.DYNAMIC_INDEX, " in the size arguments"
        )
        verify_dynamic_index_list(
            static_offsets, self.offsets, self.DYNAMIC_INDEX, " in the offset arguments"
        )
        verify_dynamic_index_list(
            static_strides, self.strides, self.DYNAMIC_INDEX, " in the stride arguments"
        )

    def __init__(
        self,
        source: SSAValue | Operation,
        offsets: Sequence[SSAValue],
        sizes: Sequence[SSAValue],
        strides: Sequence[SSAValue],
        static_offsets: Sequence[int] | DenseArrayBase,
        static_sizes: Sequence[int] | DenseArrayBase,
        static_strides: Sequence[int] | DenseArrayBase,
        result_type: Attribute,
    ):
        if not isinstance(static_offsets, DenseArrayBase):
            static_offsets = DenseArrayBase.from_list(i64, static_offsets)
        if not isinstance(static_sizes, DenseArrayBase):
            static_sizes = DenseArrayBase.from_list(i64, static_sizes)
        if not isinstance(static_strides, DenseArrayBase):
            static_strides = DenseArrayBase.from_list(i64, static_strides)
        super().__init__(
            operands=[source, offsets, sizes, strides],
            result_types=[result_type],
            properties={
                "static_offsets": static_offsets,
                "static_sizes": static_sizes,
                "static_strides": static_strides,
            },
        )

    @staticmethod
    def get(
        source: SSAValue,
        offsets: Sequence[SSAValue | int],
        sizes: Sequence[SSAValue | int],
        strides: Sequence[SSAValue | int],
        result_type: Attribute,
    ) -> SubviewOp:
        static_offsets, dyn_offsets = split_dynamic_index_list(
            offsets, SubviewOp.DYNAMIC_INDEX
        )
        static_sizes, dyn_sizes = split_dynamic_index_list(
            sizes, SubviewOp.DYNAMIC_INDEX
        )
        static_strides, dyn_strides = split_dynamic_index_list(
            strides, SubviewOp.DYNAMIC_INDEX
        )

        return SubviewOp(
            source,
            dyn_offsets,
            dyn_sizes,
            dyn_strides,
            static_offsets,
            static_sizes,
            static_strides,
            result_type,
        )

    @staticmethod
    def from_static_parameters(
        source: SSAValue | Operation,
        source_type: MemRefType,
        offsets: Sequence[int],
        sizes: Sequence[int],
        strides: Sequence[int],
        reduce_rank: bool = False,
    ) -> SubviewOp:
        source = SSAValue.get(source)

        source_shape = source_type.get_shape()
        source_offset = 0
        source_strides = [1]
        for input_size in reversed(source_shape[1:]):
            source_strides.insert(0, source_strides[0] * input_size)
        if isinstance(source_type.layout, StridedLayoutAttr):
            if isinstance(source_type.layout.offset, IntAttr):
                source_offset = source_type.layout.offset.data
            if isa(source_type.layout.strides, ArrayAttr[IntAttr]):
                source_strides = [s.data for s in source_type.layout.strides]

        layout_strides = [a * b for (a, b) in zip(strides, source_strides)]

        layout_offset = (
            sum(stride * offset for stride, offset in zip(source_strides, offsets))
            + source_offset
        )

        if reduce_rank:
            composed_strides = layout_strides
            layout_strides: list[int] = []
            result_sizes: list[int] = []

            for stride, size in zip(composed_strides, sizes):
                if size == 1:
                    continue
                layout_strides.append(stride)
                result_sizes.append(size)

        else:
            result_sizes = list(sizes)

        layout = StridedLayoutAttr(layout_strides, layout_offset)

        return_type = MemRefType(
            source_type.element_type,
            result_sizes,
            layout,
            source_type.memory_space,
        )

        return SubviewOp(
            source,
            (),
            (),
            (),
            DenseArrayBase.from_list(i64, offsets),
            DenseArrayBase.from_list(i64, sizes),
            DenseArrayBase.from_list(i64, strides),
            return_type,
        )


@irdl_op_definition
class CastOp(IRDLOperation):
    name = "memref.cast"

    source = operand_def(base(MemRefType) | base(UnrankedMemRefType))
    dest = result_def(base(MemRefType) | base(UnrankedMemRefType))

    traits = traits_def(NoMemoryEffect())

    @staticmethod
    def get(
        source: SSAValue | Operation,
        type: MemRefType | UnrankedMemRefType,
    ):
        return CastOp.build(operands=[source], result_types=[type])


@irdl_op_definition
class MemorySpaceCastOp(IRDLOperation):
    name = "memref.memory_space_cast"

    source = operand_def(base(MemRefType) | base(UnrankedMemRefType))
    dest = result_def(base(MemRefType) | base(UnrankedMemRefType))

    traits = traits_def(NoMemoryEffect())

    def __init__(
        self,
        source: SSAValue | Operation,
        dest: MemRefType | UnrankedMemRefType,
    ):
        super().__init__(operands=[source], result_types=[dest])

    @staticmethod
    def from_type_and_target_space(
        source: SSAValue | Operation,
        type: MemRefType,
        dest_memory_space: Attribute,
    ) -> MemorySpaceCastOp:
        dest = MemRefType(
            type.get_element_type(),
            shape=type.get_shape(),
            layout=type.layout,
            memory_space=dest_memory_space,
        )
        return MemorySpaceCastOp(source, dest)

    def verify_(self) -> None:
        source = cast(MemRefType, self.source.type)
        dest = cast(MemRefType, self.dest.type)
        if source.get_shape() != dest.get_shape():
            raise VerifyException(
                "Expected source and destination to have the same shape."
            )
        if source.get_element_type() != dest.get_element_type():
            raise VerifyException(
                "Expected source and destination to have the same element type."
            )


@irdl_op_definition
class ReinterpretCastOp(IRDLOperation):
    DYNAMIC_INDEX: ClassVar[int] = -9223372036854775808

    name = "memref.reinterpret_cast"

    source = operand_def(MemRefType)

    offsets = var_operand_def(IndexType)
    sizes = var_operand_def(IndexType)
    strides = var_operand_def(IndexType)

    static_offsets = prop_def(DenseArrayBase.constr(i64))
    static_sizes = prop_def(DenseArrayBase.constr(i64))
    static_strides = prop_def(DenseArrayBase.constr(i64))

    result = result_def(MemRefType)

    traits = traits_def(NoMemoryEffect())

    irdl_options = [AttrSizedOperandSegments(as_property=True)]

    assembly_format = (
        "$source `to` `offset` `` `:`"
        "custom<DynamicIndexList>($offsets, $static_offsets)"
        "`` `,` `sizes` `` `:`"
        "custom<DynamicIndexList>($sizes, $static_sizes)"
        "`` `,` `strides` `` `:`"
        "custom<DynamicIndexList>($strides, $static_strides)"
        "attr-dict `:` type($source) `to` type($result)"
    )

    custom_directives = (DynamicIndexList,)

    def __init__(
        self,
        source: SSAValue | Operation,
        offsets: Sequence[SSAValue],
        sizes: Sequence[SSAValue],
        strides: Sequence[SSAValue],
        static_offsets: Sequence[int] | DenseArrayBase,
        static_sizes: Sequence[int] | DenseArrayBase,
        static_strides: Sequence[int] | DenseArrayBase,
        result_type: Attribute,
    ):
        if not isinstance(static_offsets, DenseArrayBase):
            static_offsets = DenseArrayBase.from_list(i64, static_offsets)
        if not isinstance(static_sizes, DenseArrayBase):
            static_sizes = DenseArrayBase.from_list(i64, static_sizes)
        if not isinstance(static_strides, DenseArrayBase):
            static_strides = DenseArrayBase.from_list(i64, static_strides)
        super().__init__(
            operands=[source, offsets, sizes, strides],
            result_types=[result_type],
            properties={
                "static_offsets": static_offsets,
                "static_sizes": static_sizes,
                "static_strides": static_strides,
            },
        )

    @staticmethod
    def from_dynamic(
        source: SSAValue,
        offsets: Sequence[SSAValue | int],
        sizes: Sequence[SSAValue | int],
        strides: Sequence[SSAValue | int],
        result_type: Attribute,
    ):
        """
        Construct a `ReinterpretCastOp` from dynamic offsets, sizes, and strides.
        """
        static_offsets, dyn_offsets = split_dynamic_index_list(
            offsets, ReinterpretCastOp.DYNAMIC_INDEX
        )
        static_sizes, dyn_sizes = split_dynamic_index_list(
            sizes, ReinterpretCastOp.DYNAMIC_INDEX
        )
        static_strides, dyn_strides = split_dynamic_index_list(
            strides, ReinterpretCastOp.DYNAMIC_INDEX
        )

        return ReinterpretCastOp(
            source,
            dyn_offsets,
            dyn_sizes,
            dyn_strides,
            static_offsets,
            static_sizes,
            static_strides,
            result_type,
        )

    def verify_(self):
        static_offsets = self.static_offsets.get_values()
        static_sizes = self.static_sizes.get_values()
        static_strides = self.static_strides.get_values()

        verify_dynamic_index_list(
            static_sizes, self.sizes, self.DYNAMIC_INDEX, " in the size arguments"
        )
        verify_dynamic_index_list(
            static_offsets, self.offsets, self.DYNAMIC_INDEX, " in the offset arguments"
        )
        verify_dynamic_index_list(
            static_strides, self.strides, self.DYNAMIC_INDEX, " in the stride arguments"
        )

        assert isa(self.source.type, MemRefType)
        assert isa(self.result.type, MemRefType)

        if len(self.result.type.shape) != len(self.static_sizes):
            raise VerifyException(
                f"Expected {len(self.source.type.shape)} size values but got {len(self.static_sizes)}"
            )

        # validate sizes
        for dim, (actual, expected) in enumerate(
            zip(
                self.result.type.get_shape(),
                self.static_sizes.get_values(),
                strict=True,
            )
        ):
            if expected == ReinterpretCastOp.DYNAMIC_INDEX and actual != -1:
                raise VerifyException(
                    f"Expected result type with dynamic size instead of {actual} in dim = {dim}"
                )
            elif expected != ReinterpretCastOp.DYNAMIC_INDEX and expected != actual:
                raise VerifyException(
                    f"Expected result type with size = {expected} instead of {actual} in dim = {dim}"
                )


@irdl_op_definition
class DmaStartOp(IRDLOperation):
    name = "memref.dma_start"

    src = operand_def(MemRefType)
    src_indices = var_operand_def(IndexType)

    dest = operand_def(MemRefType)
    dest_indices = var_operand_def(IndexType)

    num_elements = operand_def(IndexType)

    tag = operand_def(MemRefType[IntegerType])
    tag_indices = var_operand_def(IndexType)

    traits = traits_def(MemoryWriteEffect(), MemoryReadEffect())

    irdl_options = [AttrSizedOperandSegments()]

    @staticmethod
    def get(
        src: SSAValue | Operation,
        src_indices: Sequence[SSAValue | Operation],
        dest: SSAValue | Operation,
        dest_indices: Sequence[SSAValue | Operation],
        num_elements: SSAValue | Operation,
        tag: SSAValue | Operation,
        tag_indices: Sequence[SSAValue | Operation],
    ):
        return DmaStartOp.build(
            operands=[
                src,
                src_indices,
                dest,
                dest_indices,
                num_elements,
                tag,
                tag_indices,
            ]
        )

    def verify_(self) -> None:
        assert isa(self.src.type, MemRefType)
        assert isa(self.dest.type, MemRefType)
        assert isa(self.tag.type, MemRefType[IntegerType])

        if len(self.src.type.shape) != len(self.src_indices):
            raise VerifyException(
                f"Expected {len(self.src.type.shape)} source indices (because of shape of src memref)"
            )

        if len(self.dest.type.shape) != len(self.dest_indices):
            raise VerifyException(
                f"Expected {len(self.dest.type.shape)} dest indices (because of shape of dest memref)"
            )

        if len(self.tag.type.shape) != len(self.tag_indices):
            raise VerifyException(
                f"Expected {len(self.tag.type.shape)} tag indices (because of shape of tag memref)"
            )

        if self.tag.type.element_type != i32:
            raise VerifyException("Expected tag to be a memref of i32")

        if self.dest.type.memory_space == self.src.type.memory_space:
            raise VerifyException("Source and dest must have different memory spaces!")


@irdl_op_definition
class DmaWaitOp(IRDLOperation):
    name = "memref.dma_wait"

    tag = operand_def(MemRefType)
    tag_indices = var_operand_def(IndexType)

    num_elements = operand_def(IndexType)

    traits = traits_def(MemoryWriteEffect(), MemoryReadEffect())

    @staticmethod
    def get(
        tag: SSAValue | Operation,
        tag_indices: Sequence[SSAValue | Operation],
        num_elements: SSAValue | Operation,
    ):
        return DmaWaitOp.build(
            operands=[
                tag,
                tag_indices,
                num_elements,
            ]
        )

    def verify_(self) -> None:
        assert isa(self.tag.type, MemRefType)

        if len(self.tag.type.shape) != len(self.tag_indices):
            raise VerifyException(
                f"Expected {len(self.tag.type.shape)} tag indices because of shape of tag memref"
            )

        if self.tag.type.element_type != i32:
            raise VerifyException("Expected tag to be a memref of i32")


@irdl_op_definition
class CopyOp(IRDLOperation):
    name = "memref.copy"
    source = operand_def(MemRefType)
    destination = operand_def(MemRefType)

    traits = traits_def(MemoryWriteEffect(), MemoryReadEffect())

    def __init__(self, source: SSAValue | Operation, destination: SSAValue | Operation):
        super().__init__(operands=[source, destination])

    def verify_(self) -> None:
        source = cast(MemRefType, self.source.type)
        destination = cast(MemRefType, self.destination.type)
        if source.get_shape() != destination.get_shape():
            raise VerifyException(
                "Expected source and destination to have the same shape."
            )
        if source.get_element_type() != destination.get_element_type():
            raise VerifyException(
                "Expected source and destination to have the same element type."
            )


MemRef = Dialect(
    "memref",
    [
        LoadOp,
        StoreOp,
        AllocOp,
        AllocaOp,
        AllocaScopeOp,
        AllocaScopeReturnOp,
        AtomicRMWOp,
        CopyOp,
        CollapseShapeOp,
        ExpandShapeOp,
        DeallocOp,
        GetGlobalOp,
        GlobalOp,
        DimOp,
        ExtractStridedMetaDataOp,
        ExtractAlignedPointerAsIndexOp,
        SubviewOp,
        CastOp,
        MemorySpaceCastOp,
        ReinterpretCastOp,
        DmaStartOp,
        DmaWaitOp,
        RankOp,
    ],
    [],
)



from collections.abc import Generator
from dataclasses import dataclass
from itertools import chain
from typing import Any, cast

from typing_extensions import TypeVar

from xdsl.context import Context
from xdsl.dialects import builtin
from xdsl.dialects.experimental.dmp import SwapOp
from xdsl.dialects.stencil import (
    AccessOp,
    AllocOp,
    ApplyOp,
    BufferOp,
    CombineOp,
    DynAccessOp,
    FieldType,
    IndexAttr,
    LoadOp,
    ReturnOp,
    StencilBoundsAttr,
    StoreOp,
    TempType,
)
from xdsl.ir import (
    Attribute,
    Block,
    Operation,
    Region,
    SSAValue,
)
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.traits import MemoryEffectKind, get_effects
from xdsl.transforms.canonicalization_patterns.stencil import ApplyUnusedResults
from xdsl.transforms.dead_code_elimination import RemoveUnusedOperations
from xdsl.utils.hints import isa

_TypeElement = TypeVar("_TypeElement", bound=Attribute)


def field_from_temp(temp: TempType[_TypeElement]) -> FieldType[_TypeElement]:
    return FieldType[_TypeElement].new(temp.parameters)


def might_effect(
    operation: Operation, effects: set[MemoryEffectKind], value: SSAValue
) -> bool:
    """
    Return True if the operation might have any of the given effects on the given value.
    """
    op_effects = get_effects(operation)
    return op_effects is None or any(
        e.kind in effects and e.value in (None, value) for e in op_effects
    )


class ApplyBufferizePattern(RewritePattern):
    """
    Naive partial `stencil.apply` bufferization.

    Just replace all temp arguments with the field result of a stencil.buffer on them, meaning
    "The buffer those value are allocated to".

    Example:
    ```mlir
    %out = stencil.apply(%0 = %in : !stencil.temp<[0,32]xf64>) -> (!stencil.temp<[0,32]>xf64) {
        // [...]
    }
    ```
    yields:
    ```mlir
    %in_buf = stencil.buffer %in : !stencil.temp<[0,32]xf64> -> !stencil.field<[0,32]xf64>
    stencil.apply(%0 = %in_buf : !stencil.field<[0,32]>xf64) outs (%out_buf : !stencil.field<[0,32]>xf64) {
        // [...]
    }
    ```
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ApplyOp, rewriter: PatternRewriter):
        if all(not isinstance(o.type, TempType) for o in op.args):
            return

        bounds = op.get_bounds()

        args = [
            (
                BufferOp.create(
                    operands=[o],
                    result_types=[field_from_temp(o.type)],
                )
                if isa(o.type, TempType[Attribute])
                else o
            )
            for o in op.args
        ]

        new = ApplyOp(
            operands=[args, op.dest],
            regions=[op.detach_region(0)],
            result_types=[op.res.types],
            properties={"bounds": bounds},
        )

        rewriter.replace_matched_op(
            [*(o for o in args if isinstance(o, Operation)), new]
        )


def walk_from(a: Operation) -> Generator[Operation, Any, None]:
    """
    Walk through all operations recursively inside a or its block.
    """
    while True:
        yield from a.walk()
        if a.next_op is None:
            break
        a = a.next_op


def walk_from_to(a: Operation, b: Operation, *, inclusive: bool = False):
    """
    Walk through all operations recursively inside a or its block, until b is met, if
    ever.
    """
    for o in walk_from(a):
        if o == b:
            if inclusive:
                yield o
            return
        yield o


def is_inplace(apply: ApplyOp, field: SSAValue):
    """
    Check if the passed `stencil.apply` has any non-zero offset access to the passed
    `stencil.field`.
    """
    # Get all block arguments matching this field
    field_args = set(
        apply.region.block.args[i] for (i, a) in enumerate(apply.args) if a is field
    )
    # Is there any non-zero access on those arguments?
    return not any(
        access
        for access in apply.walk()
        if isinstance(access, AccessOp)
        and access.temp in field_args
        and any(o != 0 for o in access.offset)
        or isinstance(access, DynAccessOp)
        and access.temp in field_args
        and any(o != 0 for o in chain(access.lb, access.ub))
    )


class LoadBufferFoldPattern(RewritePattern):
    """
    Fold a reference-semantic `stencil.buffer` of a `stencil.load` to the underlying
    field if safe.

    Example:
    ```mlir
    %temp = stencil.load %field : !stencil.field<[-2,34]> -> !stencil.temp<[0,32]>
    // [... No changes on %field]
    %temp_f = stencil.buffer %temp : !stencil.temp<[0,32]> -> !stencil.field<[0,32]>
    // [... No changes on %field]
    // Last use of temp_f
    ```
    yields:
    ```mlir
    // Will be simplified away or folded again
    %temp = stencil.load %field : !stencil.field<[-2,34]> -> !stencil.temp<[0,32]>
    // [... %temp_f replaced by %field]
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: BufferOp, rewriter: PatternRewriter):
        # If this is a value-semantic buffer, we can't fold it
        if not isinstance(op.res.type, FieldType):
            return

        temp = op.temp
        load = temp.owner
        # We are interested in folding buffers of loaded values
        if not isinstance(load, LoadOp):
            return

        underlying = load.field

        # TODO: further analysis
        # For now, only handle usages in the same block
        uses = tuple(op.res.uses)
        block = op.parent
        if not block or any(use.operation.parent is not block for use in uses):
            return
        last_user = max(
            uses, key=lambda u: block.get_operation_index(u.operation)
        ).operation

        effecting = [
            o
            for o in walk_from_to(load, last_user, inclusive=True)
            if might_effect(o, {MemoryEffectKind.WRITE}, underlying)
        ]

        # If the last effecting op is a stencil, handle the safe inplace case
        if (
            effecting
            and isinstance(effecting[-1], ApplyOp)
            and is_inplace(effecting[-1], op.res)
        ):
            effecting.pop()
        if effecting:
            return

        rewriter.replace_matched_op(new_ops=[], new_results=[underlying])


class ApplyStoreFoldPattern(RewritePattern):
    """
    Fold stores of applys result

    Example:
    ```mlir
    %temp = stencil.apply() -> (!stencil.temp<[0,32]>) {
        // [...]
    }
    // [... %dest not read]
    stencil.store %temp to %dest (<[0], [32]>) : !stencil.temp<[0,32]> to !stencil.field<[-2,34]>
    ```
    yields:
    ```mlir
    // Outputs on dest directly
    stencil.apply() outs (%dest : !stencil.field<[-2,34]>) {
        // [...]
    }
    ```
    """

    @staticmethod
    def is_dest_safe(apply: ApplyOp, store: StoreOp) -> bool:
        # Check that the destination is not used between the apply and store.
        dest = store.field
        effecting = [
            o
            for o in walk_from_to(apply, store)
            if might_effect(o, {MemoryEffectKind.READ, MemoryEffectKind.WRITE}, dest)
        ]
        return not effecting

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ApplyOp, rewriter: PatternRewriter):
        apply = op
        for temp_index, stored in enumerate(op.res):
            # We are looking for a result that is stored and foldable
            stores = [
                use.operation
                for use in stored.uses
                if isinstance(use.operation, StoreOp)
                and self.is_dest_safe(apply, use.operation)
            ]
            if not stores:
                continue

            bounds = apply.get_bounds()
            if not isinstance(bounds, StencilBoundsAttr):
                raise ValueError(
                    "Stencil shape inference must be ran before bufferization."
                )

            new_apply = ApplyOp.build(
                # We add new destinations for each store of the removed result
                operands=[
                    apply.args,
                    (*apply.dest, *(store.field for store in stores)),
                ],
                # We only remove the considered result
                result_types=[
                    [
                        r.type
                        for r in apply.results[:temp_index]
                        + apply.results[temp_index + 1 :]
                    ]
                ],
                properties=apply.properties.copy() | {"bounds": bounds},
                attributes=apply.attributes.copy(),
                # The block signature is the same
                regions=[
                    Region(Block(arg_types=[SSAValue.get(a).type for a in apply.args])),
                ],
            )

            # The body is the same
            rewriter.inline_block(
                apply.region.block,
                InsertPoint.at_start(new_apply.region.block),
                new_apply.region.block.args,
            )

            # We swap the return's operand order, to make sure the order still matches destinations
            # after bufferization
            old_return = new_apply.region.block.last_op
            assert isinstance(old_return, ReturnOp)
            uf = old_return.unroll_factor
            new_return_args = list(
                old_return.arg[: uf * temp_index]
                + old_return.arg[uf * (temp_index + 1) :]
                + old_return.arg[uf * temp_index : uf * (temp_index + 1)] * len(stores)
            )
            new_return = ReturnOp.create(
                operands=new_return_args,
                properties=old_return.properties.copy(),
                attributes=old_return.attributes.copy(),
            )
            rewriter.replace_op(old_return, new_return)

            # Create a load of a destination, for any other user of the result
            load = LoadOp.get(stores[0].field, bounds.lb, bounds.ub)

            rewriter.replace_matched_op(
                [new_apply, load],
                new_apply.results[:temp_index]
                + (load.res,)
                + new_apply.results[temp_index:],
            )
            for store in stores:
                rewriter.erase_op(store)
            return


@dataclass(frozen=True)
class UpdateApplyArgs(RewritePattern):
    """
    Stencil bufferization will often replace a temporary apply's argument with a wider
    one.
    This pattern simply updates block arguments accordingly.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ApplyOp, rewriter: PatternRewriter):
        new_arg_types = op.args.types
        if new_arg_types == op.region.block.arg_types:
            return

        new_block = Block(arg_types=new_arg_types)
        new_apply = ApplyOp.create(
            operands=op.operands,
            result_types=op.result_types,
            properties=op.properties.copy(),
            attributes=op.attributes.copy(),
            regions=[Region(new_block)],
        )

        rewriter.inline_block(
            op.region.block, InsertPoint.at_start(new_block), new_block.args
        )

        rewriter.replace_matched_op(new_apply)


@dataclass(frozen=True)
class BufferAlloc(RewritePattern):
    """
    Replace a value semantic `stencil.buffer` by a load from an allocated field, after
    a store of the input values on it.

    This matches the orginal dialect's lowering for this operation.

    Example:
    ```mlir
    // [...]
    %forward = stencil.buffer %in : !stencil.temp<[0,32]> -> !stencil.temp<[0,32]>
    // [...]
    ```
    yields:
    ```mlir
    %alloc = stencil.alloc : !stencil.field<[0,32]>xf64
    // [...]
    // This should be folded in the above computation
    stencil.store %in to %alloc (<[0], [32]>) : !stencil.temp<[0,32]> to !stencil.field<[0,32]>
    // This should be folded in the below computation
    %forward = stencil.load %alloc : !stencil.field<[0,32]>xf64 -> !stencil.temp<[0,32]>
    // [...]
    ```
    """  # noqa: E501

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: BufferOp, rewriter: PatternRewriter):
        # If it's not a value-semantic buffer, we let other patterns work on it.
        if not isinstance(op.res.type, TempType):
            return

        temp_t = cast(TempType[Attribute], op.temp.type)
        if not isinstance(temp_t.bounds, StencilBoundsAttr):
            raise ValueError(
                "Stencil shape inference must be ran before bufferization."
            )
        alloc = AllocOp(result_types=[field_from_temp(temp_t)])
        rewriter.insert_op(alloc, InsertPoint.at_start(cast(Block, op.parent)))

        rewriter.replace_matched_op(
            new_ops=[
                StoreOp.get(op.temp, alloc.field, temp_t.bounds),
                LoadOp.get(alloc.field, temp_t.bounds.lb, temp_t.bounds.ub),
            ]
        )


@dataclass(frozen=True)
class CombineStoreFold(RewritePattern):
    """
    A stored combine result is folded into stores of the matching operand in the
    destination field.

    Example:
    ```mlir
    %res1, %res2 = stencil.combine 1 at 11 lower = (%0 : !stencil.temp<[0,16]xf64>) upper = (%1 : !stencil.temp<[16,32]xf64>) lowerext = (%2 : !stencil.temp<[0,16]xf64>): !stencil.temp<[0,32]xf64>, !stencil.temp<[0,32]xf64>
    stencil.store %res1 to %dest1 (<[0], [32]>) : !stencil.temp<[0,32]xf64> to !stencil.field<[-2,34]xf64>
    stencil.store %res2 to %dest2 (<[0], [32]>) : !stencil.temp<[0,32]xf64> to !stencil.field<[-2,34]xf64>
    ```
    yields:
    ```mlir
    stencil.store %0 to %dest1 (<[0], [16]>) : !stencil.temp<[0,16]xf64> to !stencil.field<[-2,34]xf64>
    stencil.store %1 to %dest1 (<[16], [32]>) : !stencil.temp<[16,32]xf64> to !stencil.field<[-2,34]xf64>
    stencil.store %2 to %dest2 (<[0], [16]>) : !stencil.temp<[0,16]xf64> to !stencil.field<[-2,34]xf64>
    ```
    """  # noqa: E501

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: CombineOp, rewriter: PatternRewriter):
        for i, r in enumerate(op.results):
            if not r.has_one_use():
                continue
            store = next(iter(r.uses)).operation
            if not isinstance(store, StoreOp):
                continue

            new_lower = op.lower
            new_upper = op.upper
            new_lowerext = op.lowerext
            new_upperext = op.upperext
            new_results_types = list(op.result_types)
            new_results_types.pop(i)

            bounds = cast(StencilBoundsAttr, cast(TempType[Attribute], r.type).bounds)
            newub = list(bounds.ub)
            newub[op.dim.value.data] = op.index.value.data
            lower_bounds = StencilBoundsAttr.new((bounds.lb, IndexAttr.get(*newub)))
            newlb = list(bounds.lb)
            newlb[op.dim.value.data] = op.index.value.data
            upper_bounds = StencilBoundsAttr.new((IndexAttr.get(*newlb), bounds.ub))

            rewriter.erase_op(store)

            # If it corresponds to a lower/upper result
            if i < len(op.lower):
                new_lower = op.lower[:i] + op.lower[i + 1 :]
                new_upper = op.upper[:i] + op.upper[i + 1 :]
                rewriter.insert_op(
                    (
                        StoreOp.get(
                            op.lower[i],
                            store.field,
                            lower_bounds,
                        ),
                        StoreOp.get(
                            op.upper[i],
                            store.field,
                            upper_bounds,
                        ),
                    ),
                    InsertPoint.before(op),
                )
            # If it corresponds to a lowerext result
            elif i < len(op.lower) + len(op.lowerext):
                new_lowerext = (
                    op.lowerext[: i - len(op.lower)]
                    + op.lowerext[i - len(op.lower) + 1 :]
                )
                rewriter.insert_op(
                    (
                        StoreOp.get(
                            op.lower[i],
                            store.field,
                            lower_bounds,
                        ),
                        StoreOp.get(
                            op.upper[i],
                            store.field,
                            upper_bounds,
                        ),
                    ),
                    InsertPoint.before(op),
                )
            else:
                new_upperext = (
                    op.upperext[: i - len(op.lower) - len(op.lowerext)]
                    + op.upperext[i - len(op.lower) - len(op.lowerext) + 1 :]
                )
                rewriter.insert_op(
                    (
                        StoreOp.get(
                            op.lower[i],
                            store.field,
                            lower_bounds,
                        ),
                        StoreOp.get(
                            op.upper[i],
                            store.field,
                            upper_bounds,
                        ),
                    ),
                    InsertPoint.before(op),
                )

            new_combine = CombineOp(
                operands=[new_lower, new_upper, new_lowerext, new_upperext],
                result_types=[new_results_types],
                attributes=op.attributes.copy(),
                properties=op.properties.copy(),
            )
            rewriter.replace_matched_op(
                new_combine,
                new_results=new_combine.results[:i] + (None,) + new_combine.results[i:],
            )
            return


class SwapBufferize(RewritePattern):
    """
    Bufferize a dmp.swap operation.

    NB: This should most likely consider a shared pass following canonicalize and
    shape-inference.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: SwapOp, rewriter: PatternRewriter):
        temp = op.input_stencil

        if not isa(temp_t := temp.type, TempType[Attribute]):
            return

        load = temp.owner
        if not isinstance(load, LoadOp):
            return

        buffer = BufferOp.create(
            operands=[temp], result_types=[field_from_temp(temp_t)]
        )
        new_swap = SwapOp.get(buffer.res, op.strategy)
        new_swap.swaps = op.swaps
        load = LoadOp(operands=[buffer.res], result_types=[temp_t])

        rewriter.replace_matched_op(
            new_ops=[buffer, new_swap, load],
        )


@dataclass(frozen=True)
class StencilBufferize(ModulePass):
    """
    Bufferize the stencil dialect, i.e., try to fold all loads, sotres, buffer and
    combines, and to output stencils working directly on buffers (fields) with
    hopefully few allocations.
    """

    name = "stencil-bufferize"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        walker = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    UpdateApplyArgs(),
                    ApplyBufferizePattern(),
                    BufferAlloc(),
                    CombineStoreFold(),
                    LoadBufferFoldPattern(),
                    ApplyStoreFoldPattern(),
                    RemoveUnusedOperations(),
                    ApplyUnusedResults(),
                    SwapBufferize(),
                ]
            ),
            apply_recursively=True,
        )
        walker.rewrite_module(op)



import itertools
from collections import deque
from collections.abc import Iterable
from dataclasses import dataclass, field
from warnings import warn

from xdsl.context import Context
from xdsl.dialects import builtin
from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import Use
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)


def _try_remove_cast_chain(
    op: builtin.UnrealizedConversionCastOp,
    rewriter: PatternRewriter,
    warn_on_failure: bool,
):
    # casts that either have no uses or have at least
    # one user that isn't an unrealized cast.
    finalized_casts = list[builtin.UnrealizedConversionCastOp]()

    # casts with only other unrealized casts as users.
    pending_casts = deque[builtin.UnrealizedConversionCastOp]()

    # casts that neeed to be visited in the DFS traversal of the use-def chain.
    casts_to_visit = [op]

    def gen_all_uses_cast(
        node: builtin.UnrealizedConversionCastOp,
    ) -> Iterable[Use]:
        for result in node.results:
            yield from result.uses

    # DFS traversal of the use-def chain.
    while casts_to_visit:
        cast = casts_to_visit.pop()

        is_live = False
        has_any_uses = False

        for use in gen_all_uses_cast(cast):
            if isinstance(use.operation, builtin.UnrealizedConversionCastOp):
                # early check, it's definitely not an unifiable cast
                if any(
                    r != i
                    for r, i in itertools.zip_longest(
                        use.operation.inputs, cast.results
                    )
                ):
                    if warn_on_failure:
                        warn(
                            f"Unable to remove cast {cast} because "
                            "it is not unifiable with its uses"
                        )
                    return
                casts_to_visit.append(use.operation)
            else:
                is_live = True
            has_any_uses = True

        # check that either there is a trivial cycle we can remove
        # because types are homogeneous (e.g. {A -> B, B -> A})
        # otherwise it means the cast is not unifiable with its uses
        assert len(cast.results) == len(op.inputs)
        has_trivial_cycle = cast.result_types == op.inputs.types
        if is_live and not has_trivial_cycle:
            if warn_on_failure:
                warn(
                    "Unable to remove cast "
                    f"{cast} because it is not unifiable "
                    "with its uses"
                )
            return

        if not has_any_uses or is_live:
            finalized_casts.append(cast)
        else:
            pending_casts.appendleft(cast)

    for cast in filter(lambda c: c.parent is not None, finalized_casts):
        # replace the uses of this cast by the inputs of the root cast
        rewriter.replace_op(cast, [], op.inputs)

    for cast in filter(lambda c: c.parent is not None, pending_casts):
        # remove other casts in the chain in the right order
        rewriter.erase_op(cast)


@dataclass
class ReconcileUnrealizedCastsPattern(RewritePattern):
    """
    Removes the chains of `builtin.unrealized_conversion_cast` operations
    that are no longer necessary and that start with the matched
    `builtin.unrealized_conversion_cast`.
    """

    warn_on_failure: bool = field(default=False, kw_only=True)

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: builtin.UnrealizedConversionCastOp, rewriter: PatternRewriter
    ):
        _try_remove_cast_chain(op, rewriter, self.warn_on_failure)


def reconcile_unrealized_casts(module: ModuleOp, *, warn_on_failure: bool = True):
    """
    Removes all `builtin.unrealized_conversion_cast` operations
    that are not needed anymore in a module.
    """

    PatternRewriteWalker(
        ReconcileUnrealizedCastsPattern(warn_on_failure=warn_on_failure)
    ).rewrite_module(module)


class ReconcileUnrealizedCastsPass(ModulePass):
    name = "reconcile-unrealized-casts"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        reconcile_unrealized_casts(op)



# RUN: python %s | filecheck %s

from xdsl.dialects.cmath import Cmath
from xdsl.dialects.irdl.pyrdl_to_irdl import dialect_to_irdl

print(dialect_to_irdl(Cmath, "cmath"))

# CHECK:      irdl.dialect @cmath {

# CHECK-NEXT:   irdl.attribute @complex {
# CHECK-NEXT:     %{{.*}} = irdl.any
# CHECK-NEXT:     irdl.parameters(elem: %{{.*}})
# CHECK-NEXT:   }

# CHECK-NEXT:   irdl.operation @norm {
# CHECK-NEXT:     %{{.*}} = irdl.any
# CHECK-NEXT:     irdl.operands(in: %{{.*}})
# CHECK-NEXT:     %{{.*}} = irdl.any
# CHECK-NEXT:     irdl.results(out: %{{.*}})
# CHECK-NEXT:   }

# CHECK-NEXT:   irdl.operation @mul {
# CHECK-NEXT:     %{{.*}} = irdl.any
# CHECK-NEXT:     %{{.*}} = irdl.any
# CHECK-NEXT:     irdl.operands(lhs: %{{.*}}, rhs: %{{.*}})
# CHECK-NEXT:     %{{.*}} = irdl.any
# CHECK-NEXT:     irdl.results(res: %{{.*}})
# CHECK-NEXT:   }

# CHECK-NEXT: }



from xdsl.dialects import arith, ptr
from xdsl.dialects.builtin import IntegerAttr
from xdsl.pattern_rewriter import (
    PatternRewriter,
    RewritePattern,
    op_type_rewrite_pattern,
)


class RedundantFromPtr(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ptr.FromPtrOp, rewriter: PatternRewriter, /):
        if not isinstance(origin := op.source.owner, ptr.ToPtrOp):
            return

        if origin.source.type == op.res.type:
            rewriter.replace_matched_op((), (origin.source,))


class RedundantToPtr(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ptr.ToPtrOp, rewriter: PatternRewriter, /):
        if not isinstance(origin := op.source.owner, ptr.FromPtrOp):
            return

        if origin.source.type == op.res.type:
            rewriter.replace_matched_op((), (origin.source,))


class PtrAddZero(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ptr.PtrAddOp, rewriter: PatternRewriter, /):
        if (
            (isinstance(offset_op := op.offset.owner, arith.ConstantOp))
            and isinstance(offset_op.value, IntegerAttr)
            and offset_op.value.value.data == 0
        ):
            rewriter.replace_matched_op((), (op.addr,))



"""Passes to manipulate fastmath flags in FP arith operations."""

from dataclasses import dataclass
from typing import Literal

from xdsl.dialects import arith, builtin
from xdsl.dialects.utils import FastMathFlag
from xdsl.passes import Context, ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)

_FASTMATH_NAMES_TO_ENUM = {str(member.value): member for member in FastMathFlag}


def _get_flag_list(flags: tuple[str, ...]):
    try:
        return tuple(_FASTMATH_NAMES_TO_ENUM[flag] for flag in flags)
    except KeyError as e:
        raise ValueError(f"{e} is not a valid fastmath flag.")


@dataclass
class AddArithFastMathFlags(RewritePattern):
    """Adds fastmath flags to FP binary operations from arith dialect."""

    fastmath_op_attr: arith.FastMathFlagsAttr

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self,
        op: arith.FloatingPointLikeBinaryOperation | arith.CmpfOp,
        rewriter: PatternRewriter,
    ) -> None:
        op.fastmath = self.fastmath_op_attr


@dataclass(frozen=True)
class AddArithFastMathFlagsPass(ModulePass):
    """Module pass that adds fastmath flags to FP binary operations from arith dialect.
    It currently does not preserve any existing fastmath flags that may already be part
    of the operation.
    By default (no arguments) it adds the "fast" flag.

    Arguments (all optional):

    - flags: {"fast", "none"} | list[str]: Set specific fastmath flags. Using "fast" or
      "none" enables or disables all flags, respectively.
    """

    name = "arith-add-fastmath"

    flags: Literal["fast", "none"] | tuple[str, ...] = "fast"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        if isinstance(self.flags, str):
            fm_flags = arith.FastMathFlagsAttr(self.flags)
        else:
            if "none" in self.flags or "fast" in self.flags:
                raise ValueError(
                    'f{"none" or "fast" cannot be provided along with other fastmath flags'
                )

            fm_flags = arith.FastMathFlagsAttr(_get_flag_list(self.flags))

        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    AddArithFastMathFlags(fm_flags),
                ]
            ),
            apply_recursively=False,
        ).rewrite_module(op)



from typing import Any, cast

import pytest
from textual.screen import Screen
from textual.widgets import Tree

from xdsl.backend.riscv.lowering import (
    convert_arith_to_riscv,
    convert_func_to_riscv_func,
)
from xdsl.builder import ImplicitBuilder
from xdsl.dialects import arith, func, get_all_dialects, riscv, riscv_func
from xdsl.dialects.builtin import (
    IndexType,
    IntegerAttr,
    ModuleOp,
    UnrealizedConversionCastOp,
)
from xdsl.interactive import _pasteboard
from xdsl.interactive.add_arguments_screen import AddArguments
from xdsl.interactive.app import InputApp
from xdsl.interactive.passes import (
    AvailablePass,
    get_condensed_pass_list,
    get_new_registered_context,
)
from xdsl.interactive.rewrites import get_all_possible_rewrites
from xdsl.ir import Block, Region
from xdsl.transforms import (
    get_all_passes,
    individual_rewrite,
)
from xdsl.transforms.experimental.dmp import stencil_global_to_local
from xdsl.utils.exceptions import ParseError


@pytest.mark.asyncio
async def test_inputs():
    """Test different inputs produce desired result."""
    async with InputApp(
        tuple(get_all_dialects().items()),
        tuple((p_name, p()) for p_name, p in sorted(get_all_passes().items())),
    ).run_test() as pilot:
        app = cast(InputApp, pilot.app)

        # clear preloaded code and unselect preselected pass
        app.input_text_area.clear()
        await pilot.pause()

        # Test no input
        assert app.output_text_area.text == "No input"
        assert app.current_module is None

        # Test incorrect input
        app.input_text_area.insert("dkjfd")
        await pilot.pause()
        assert (
            app.output_text_area.text
            == "<unknown>:1:6\ndkjfd\n     ^\n     Operation builtin.unregistered does not have a custom format.\n"
        )

        assert isinstance(app.current_module, ParseError)
        assert (
            str(app.current_module)
            == "<unknown>:1:6\ndkjfd\n     ^\n     Operation builtin.unregistered does not have a custom format.\n"
        )

        # Test corect input
        app.input_text_area.clear()
        app.input_text_area.insert(
            """
        func.func @hello(%n : index) -> index {
          %two = arith.constant 2 : index
          %res = arith.muli %n, %two : index
          func.return %res : index
        }
        """
        )
        await pilot.pause()
        assert (
            app.output_text_area.text
            == """builtin.module {
  func.func @hello(%n : index) -> index {
    %two = arith.constant 2 : index
    %res = arith.muli %n, %two : index
    func.return %res : index
  }
}
"""
        )

        index = IndexType()

        expected_module = ModuleOp(Region([Block()]))
        with ImplicitBuilder(expected_module.body):
            function = func.FuncOp("hello", ((index,), (index,)))
            with ImplicitBuilder(function.body) as (n,):
                two = arith.ConstantOp(IntegerAttr(2, index)).result
                res = arith.MuliOp(n, two)
                func.ReturnOp(res)

        assert isinstance(app.current_module, ModuleOp)
        assert app.current_module.is_structurally_equivalent(expected_module)


@pytest.mark.asyncio
async def test_buttons():
    """Test pressing keys has the desired result."""
    async with InputApp(
        tuple(get_all_dialects().items()),
        tuple((p_name, p()) for p_name, p in sorted(get_all_passes().items())),
    ).run_test() as pilot:
        app = cast(InputApp, pilot.app)

        # clear preloaded code and unselect preselected pass
        app.input_text_area.clear()

        await pilot.pause()
        app.input_text_area.insert(
            """
        func.func @hello(%n : index) -> index {
          %two = arith.constant 2 : index
          %res = arith.muli %n, %two : index
          func.return %res : index
        }
        """
        )

        # assert that the Input and Output Text Area's have changed
        await pilot.pause()
        assert (
            app.input_text_area.text
            == """
        func.func @hello(%n : index) -> index {
          %two = arith.constant 2 : index
          %res = arith.muli %n, %two : index
          func.return %res : index
        }
        """
        )
        assert (
            app.output_text_area.text
            == """builtin.module {
  func.func @hello(%n : index) -> index {
    %two = arith.constant 2 : index
    %res = arith.muli %n, %two : index
    func.return %res : index
  }
}
"""
        )

        # Test clicking the "clear input" button
        await pilot.click("#clear_input_button")

        # assert that the input text area has been cleared
        await pilot.pause()
        assert app.input_text_area.text == ""

        app.input_text_area.insert(
            """
        func.func @hello(%n : index) -> index {
          %two = arith.constant 2 : index
          %res = arith.muli %n, %two : index
          func.return %res : index
        }
        """
        )

        # Select two passes
        app.pass_pipeline = (
            *app.pass_pipeline,
            convert_func_to_riscv_func.ConvertFuncToRiscvFuncPass(),
        )

        app.pass_pipeline = (
            *app.pass_pipeline,
            convert_arith_to_riscv.ConvertArithToRiscvPass(),
        )

        # assert that pass selection affected Output Text Area
        await pilot.pause()
        assert (
            app.output_text_area.text
            == """\
builtin.module {
  riscv_func.func public @hello(%n : !riscv.reg<a0>) -> !riscv.reg<a0> attributes {p2align = 2 : i8} {
    %0 = riscv.mv %n : (!riscv.reg<a0>) -> !riscv.reg
    %n_1 = builtin.unrealized_conversion_cast %0 : !riscv.reg to index
    %two = riscv.li 2 : !riscv.reg
    %two_1 = builtin.unrealized_conversion_cast %two : !riscv.reg to index
    %res = builtin.unrealized_conversion_cast %n_1 : index to !riscv.reg
    %res_1 = builtin.unrealized_conversion_cast %two_1 : index to !riscv.reg
    %res_2 = riscv.mul %res, %res_1 : (!riscv.reg, !riscv.reg) -> !riscv.reg
    %res_3 = builtin.unrealized_conversion_cast %res_2 : !riscv.reg to index
    %res_4 = builtin.unrealized_conversion_cast %res_3 : index to !riscv.reg
    %1 = riscv.mv %res_4 : (!riscv.reg) -> !riscv.reg<a0>
    riscv_func.return %1 : !riscv.reg<a0>
  }
}
"""
        )

        # Test that the current pipeline command is correctly copied
        def callback(x: str):
            assert (
                x == "xdsl-opt -p 'convert-func-to-riscv-func,convert-arith-to-riscv'"
            )

        _pasteboard._test_pyclip_callback = callback  # pyright: ignore[reportPrivateUsage]
        await pilot.click("#copy_query_button")

        current_pipeline = app.pass_pipeline
        # press "Remove Last Pass" button
        await pilot.click("#remove_last_pass_button")
        await pilot.pause()

        assert app.pass_pipeline == current_pipeline[:-1]

        assert (
            app.output_text_area.text
            == """\
builtin.module {
  riscv_func.func public @hello(%n : !riscv.reg<a0>) -> !riscv.reg<a0> attributes {p2align = 2 : i8} {
    %0 = riscv.mv %n : (!riscv.reg<a0>) -> !riscv.reg
    %n_1 = builtin.unrealized_conversion_cast %0 : !riscv.reg to index
    %two = arith.constant 2 : index
    %res = arith.muli %n_1, %two : index
    %res_1 = builtin.unrealized_conversion_cast %res : index to !riscv.reg
    %1 = riscv.mv %res_1 : (!riscv.reg) -> !riscv.reg<a0>
    riscv_func.return %1 : !riscv.reg<a0>
  }
}
"""
        )

        # press "Clear Passes" button
        await pilot.click("#clear_passes_button")

        # assert that the Output Text Area and current_module have the expected results
        await pilot.pause()
        assert app.pass_pipeline == ()
        assert (
            app.output_text_area.text
            == """builtin.module {
  func.func @hello(%n : index) -> index {
    %two = arith.constant 2 : index
    %res = arith.muli %n, %two : index
    func.return %res : index
  }
}
"""
        )

        index = IndexType()

        expected_module = ModuleOp(Region([Block()]))
        with ImplicitBuilder(expected_module.body):
            function = func.FuncOp("hello", ((index,), (index,)))
            with ImplicitBuilder(function.body) as (n,):
                n.name_hint = "n"
                two = arith.ConstantOp(IntegerAttr(2, index)).result
                two.name_hint = "two"
                res = arith.MuliOp(n, two).result
                res.name_hint = "res"
                func.ReturnOp(res)

        assert isinstance(app.current_module, ModuleOp)
        assert app.current_module.is_structurally_equivalent(expected_module)

        # assert initial state of condense_mode is False
        assert app.condense_mode is False

        # press "Condense" button
        await pilot.click("#condense_button")

        await pilot.pause()
        # assert after "Condense Button" is clicked that the state and condensed_pass list change accordingly
        assert app.condense_mode is True
        rewrites = get_all_possible_rewrites(expected_module)
        assert app.available_pass_list == get_condensed_pass_list(
            get_new_registered_context(app.all_dialects),
            expected_module,
            app.all_passes,
        ) + tuple(rewrites)

        # press "Uncondense" button
        await pilot.click("#uncondense_button")

        await pilot.pause()
        # assert after "Condense Button" is clicked that the state changes accordingly
        assert app.condense_mode is False


@pytest.mark.asyncio
async def test_rewrites():
    """Test rewrite application has the desired result."""
    async with InputApp(
        tuple(get_all_dialects().items()),
        tuple((p_name, p()) for p_name, p in sorted(get_all_passes().items())),
    ).run_test() as pilot:
        app = cast(InputApp, pilot.app)
        # clear preloaded code and unselect preselected pass
        app.input_text_area.clear()

        await pilot.pause()
        # Testing a pass
        app.input_text_area.insert(
            """
        func.func @hello(%n : i32) -> i32 {
  %c0 = arith.constant 0 : i32
  %res = arith.addi %n, %c0 : i32
  func.return %res : i32
}
        """
        )

        # press "Condense" button
        await pilot.click("#condense_button")

        addi_pass = AvailablePass(
            module_pass=individual_rewrite.ApplyIndividualRewritePass(
                3, "arith.addi", "SignlessIntegerBinaryOperationZeroOrUnitRight"
            ),
        )

        await pilot.pause()
        # assert after "Condense Button" is clicked that the state and get_condensed_pass list change accordingly
        assert app.condense_mode is True
        assert isinstance(app.current_module, ModuleOp)
        condensed_list = get_condensed_pass_list(
            get_new_registered_context(app.all_dialects),
            app.current_module,
            app.all_passes,
        ) + (addi_pass,)
        assert app.available_pass_list == condensed_list

        # Select a rewrite
        app.pass_pipeline = (
            *app.pass_pipeline,
            individual_rewrite.ApplyIndividualRewritePass(
                3, "arith.addi", "SignlessIntegerBinaryOperationZeroOrUnitRight"
            ),
        )

        # assert that pass selection affected Output Text Area
        await pilot.pause()
        assert (
            app.output_text_area.text
            == """builtin.module {
  func.func @hello(%n : i32) -> i32 {
    %c0 = arith.constant 0 : i32
    func.return %n : i32
  }
}
"""
        )


@pytest.mark.asyncio
async def test_passes():
    """Test pass application has the desired result."""
    async with InputApp(
        tuple(get_all_dialects().items()),
        tuple((p_name, p()) for p_name, p in sorted(get_all_passes().items())),
    ).run_test() as pilot:
        app = cast(InputApp, pilot.app)
        # clear preloaded code and unselect preselected pass
        app.input_text_area.clear()

        await pilot.pause()
        # Testing a pass
        app.input_text_area.insert(
            """
        func.func @hello(%n : index) -> index {
          %two = arith.constant 2 : index
          %res = arith.muli %n, %two : index
          func.return %res : index
        }
        """
        )

        # Await on test update to make sure we only update due to pass change later
        await pilot.pause()
        assert (
            app.output_text_area.text
            == """builtin.module {
  func.func @hello(%n : index) -> index {
    %two = arith.constant 2 : index
    %res = arith.muli %n, %two : index
    func.return %res : index
  }
}
"""
        )

        # Select a pass
        app.pass_pipeline = (
            *app.pass_pipeline,
            convert_func_to_riscv_func.ConvertFuncToRiscvFuncPass(),
        )
        # assert that the Output Text Area has changed accordingly
        await pilot.pause()
        assert (
            app.output_text_area.text
            == """builtin.module {
  riscv_func.func public @hello(%n : !riscv.reg<a0>) -> !riscv.reg<a0> attributes {p2align = 2 : i8} {
    %0 = riscv.mv %n : (!riscv.reg<a0>) -> !riscv.reg
    %n_1 = builtin.unrealized_conversion_cast %0 : !riscv.reg to index
    %two = arith.constant 2 : index
    %res = arith.muli %n_1, %two : index
    %res_1 = builtin.unrealized_conversion_cast %res : index to !riscv.reg
    %1 = riscv.mv %res_1 : (!riscv.reg) -> !riscv.reg<a0>
    riscv_func.return %1 : !riscv.reg<a0>
  }
}
"""
        )

        index = IndexType()
        expected_module = ModuleOp(Region([Block()]))
        with ImplicitBuilder(expected_module.body):
            function = riscv_func.FuncOp(
                "hello",
                Region([Block(arg_types=[riscv.Registers.A0])]),
                ((riscv.Registers.A0,), (riscv.Registers.A0,)),
                "public",
                p2align=2,
            )
            with ImplicitBuilder(function.body) as (n,):
                zero = riscv.MVOp(n)
                n_one = UnrealizedConversionCastOp.get([zero.rd], [index])
                two = arith.ConstantOp(IntegerAttr(2, index)).result
                res = arith.MuliOp(n_one, two)
                one = UnrealizedConversionCastOp.get(
                    [res.result], [riscv.Registers.UNALLOCATED_INT]
                )
                two_two = riscv.MVOp(one, rd=riscv.Registers.A0)
                riscv_func.ReturnOp(two_two)

        assert isinstance(app.current_module, ModuleOp)
        # Assert that the current module has been changed accordingly
        assert app.current_module.is_structurally_equivalent(expected_module)


@pytest.mark.asyncio
async def test_argument_pass_screen():
    """Test that clicking on a pass that requires passes opens a screen to specify them."""
    async with InputApp(
        tuple(get_all_dialects().items()),
        tuple((p_name, p()) for p_name, p in sorted(get_all_passes().items())),
    ).run_test() as pilot:
        app = cast(InputApp, pilot.app)
        # clear preloaded code and unselect preselected pass
        app.input_text_area.clear()

        await pilot.pause()
        # Testing a pass
        app.input_text_area.insert(
            """
        func.func @hello(%n : i32) -> i32 {
  %two = arith.constant 0 : i32
  %res = arith.addi %two, %n : i32
  func.return %res : i32
}
        """
        )
        app.passes_tree.root.expand
        await pilot.pause()

        root_children = app.passes_tree.root.children
        distribute_stencil_node = None

        for node in root_children:
            assert node.data is not None
            if node.data is stencil_global_to_local.DistributeStencilPass:
                distribute_stencil_node = node

        assert distribute_stencil_node is not None

        # Ideally, we would like to trigger the event like this:
        # `app.passes_tree.select_node(distribute_stencil_node)`
        # When running in a test, node selection does not send the expected event
        # For now, trigger the expected method directly
        app.update_pass_pipeline(Tree.NodeSelected(distribute_stencil_node))
        await pilot.pause()

        arg_screen_str: type[Screen[Any]] = AddArguments
        assert isinstance(app.screen, arg_screen_str)


@pytest.mark.asyncio
async def test_dark_mode():
    """Tests that 'd' switches between dark and light mode"""

    async with InputApp(tuple(), tuple()).run_test() as pilot:
        app = cast(InputApp, pilot.app)

        assert app.theme == "textual-dark"

        await pilot.press("d")

        assert app.theme == "textual-light"

        await pilot.press("d")

        assert app.theme == "textual-dark"


@pytest.mark.asyncio
async def test_apply_individual_rewrite():
    """Tests that using the tree to apply an individual rewrite works"""

    async with InputApp(tuple(get_all_dialects().items()), ()).run_test() as pilot:
        app = cast(InputApp, pilot.app)
        # clear preloaded code and unselect preselected pass
        app.input_text_area.clear()

        await pilot.pause()
        # Testing a pass
        app.input_text_area.insert(
            """
        func.func @hello(%n : i32) -> i32 {
  %c0 = arith.constant 0 : i32
  %res = arith.addi %c0, %n : i32
  func.return %res : i32
}
        """
        )
        app.passes_tree.root.expand()
        await pilot.click("#condense_button")

        node = None
        for n in app.passes_tree.root.children:
            if n.data == individual_rewrite.ApplyIndividualRewritePass(
                3, "arith.addi", "SignlessIntegerBinaryOperationConstantProp"
            ):
                node = n

        assert node is not None

        # manually trigger node selection
        app.passes_tree.select_node(node)

        await pilot.pause()

        assert (
            app.output_text_area.text
            == """builtin.module {
  func.func @hello(%n : i32) -> i32 {
    %c0 = arith.constant 0 : i32
    %res = arith.addi %n, %c0 : i32
    func.return %res : i32
  }
}
"""
        )

        # Apply second individual rewrite
        node = None
        for n in app.passes_tree.root.children:
            if n.data == individual_rewrite.ApplyIndividualRewritePass(
                3, "arith.addi", "SignlessIntegerBinaryOperationZeroOrUnitRight"
            ):
                node = n

        assert node is not None

        # manually trigger node selection
        app.passes_tree.select_node(node)

        await pilot.pause()

        assert (
            app.output_text_area.text
            == """builtin.module {
  func.func @hello(%n : i32) -> i32 {
    %c0 = arith.constant 0 : i32
    func.return %n : i32
  }
}
"""
        )



import re

import pytest
from typing_extensions import Self

from xdsl.backend.register_allocatable import RegisterConstraints
from xdsl.backend.riscv.register_allocation import (
    RegisterAllocatorLivenessBlockNaive,
    reg_types_by_name,
)
from xdsl.backend.riscv.register_stack import RiscvRegisterStack
from xdsl.dialects import riscv
from xdsl.dialects.test import TestOp
from xdsl.ir import SSAValue
from xdsl.irdl import IRDLOperation, irdl_op_definition, operand_def, result_def
from xdsl.utils.exceptions import DiagnosticException


def test_default_reserved_registers():
    available_registers = RiscvRegisterStack(allow_infinite=True)

    unallocated = riscv.Registers.UNALLOCATED_INT

    j = riscv.IntRegisterType.infinite_register

    assert available_registers.pop(riscv.IntRegisterType) == j(0)

    register_allocator = RegisterAllocatorLivenessBlockNaive(available_registers)

    assert not register_allocator.allocate_values_same_reg(())

    op_a = TestOp(result_types=[unallocated])
    register_allocator.allocate_values_same_reg(op_a.results)

    assert op_a.results[0].type == j(1)

    register_allocator.allocate_values_same_reg(op_a.results)

    assert op_a.results[0].type == j(1)

    op_b = TestOp(result_types=[unallocated, unallocated])

    register_allocator.allocate_values_same_reg(op_b.results)

    assert tuple(op_b.result_types) == (j(2), j(2))

    op_c = TestOp(result_types=[j(2), unallocated])

    register_allocator.allocate_values_same_reg(op_c.results)

    assert tuple(op_c.result_types) == (j(2), j(2))

    op_d = TestOp(result_types=[j(2), j(3)])

    with pytest.raises(
        DiagnosticException,
        match=re.escape(
            "Cannot allocate registers to the same register ['!riscv.reg<j_2>', '!riscv.reg<j_3>']"
        ),
    ):
        register_allocator.allocate_values_same_reg(op_d.results)

    op_e = TestOp(result_types=[j(2), j(3), unallocated])

    with pytest.raises(
        DiagnosticException,
        match=re.escape(
            "Cannot allocate registers to the same register ['!riscv.reg', '!riscv.reg<j_2>', '!riscv.reg<j_3>']"
        ),
    ):
        register_allocator.allocate_values_same_reg(op_e.results)


def test_allocate_with_inout_constraints():
    @irdl_op_definition
    class MyInstructionOp(riscv.RISCVAsmOperation, IRDLOperation):
        name = "riscv.my_instruction"

        rs0 = operand_def()
        rs1 = operand_def()
        rd0 = result_def()
        rd1 = result_def()

        @classmethod
        def get(cls, rs0: SSAValue, rs1: SSAValue, rd0: str, rd1: str) -> Self:
            return cls.build(
                operands=(rs0, rs1),
                result_types=(
                    riscv.IntRegisterType.from_name(rd0),
                    riscv.IntRegisterType.from_name(rd1),
                ),
            )

        def get_register_constraints(self) -> RegisterConstraints:
            return RegisterConstraints(
                (self.rs0,), (self.rd0,), ((self.rs1, self.rd1),)
            )

    available_registers = RiscvRegisterStack(allow_infinite=True)
    register_allocator = RegisterAllocatorLivenessBlockNaive(available_registers)

    # All new registers. The result register is reused by the allocator for the operand.
    rs0, rs1 = TestOp(
        result_types=[
            riscv.IntRegisterType.unallocated(),
            riscv.IntRegisterType.unallocated(),
        ]
    ).results
    op0 = MyInstructionOp.get(rs0, rs1, "", "")
    op0.allocate_registers(register_allocator)
    assert op0.rs0.type == riscv.IntRegisterType.infinite_register(1)
    assert op0.rs1.type == riscv.IntRegisterType.infinite_register(0)
    assert op0.rd0.type == riscv.IntRegisterType.infinite_register(1)
    assert op0.rd1.type == riscv.IntRegisterType.infinite_register(0)

    # One register reserved for inout parameter, the allocator should allocate the output
    # to the same register.
    rs0, rs1 = TestOp(
        result_types=[
            riscv.IntRegisterType.unallocated(),
            riscv.IntRegisterType.unallocated(),
        ]
    ).results
    op1 = MyInstructionOp.get(rs0, rs1, "", "a0")
    op1.allocate_registers(register_allocator)
    assert op1.rs0.type == riscv.IntRegisterType.infinite_register(2)
    assert op1.rs1.type == riscv.IntRegisterType.from_name("a0")
    assert op1.rd0.type == riscv.IntRegisterType.infinite_register(2)
    assert op1.rd1.type == riscv.IntRegisterType.from_name("a0")


def test_count_reg_types():
    a0 = riscv.Registers.A0
    a1 = riscv.Registers.A1

    fa0 = riscv.Registers.FA0

    assert reg_types_by_name([a0, a0, a1, fa0, fa0]) == {
        "riscv.reg": {"a0", "a1"},
        "riscv.freg": {"fa0"},
    }
